# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import six
import os
import shutil

from mock import patch, MagicMock, call
from productmd.rpms import Rpms

from odcs.server import conf, db
from odcs.server.models import Compose
from odcs.common.types import COMPOSE_FLAGS, COMPOSE_RESULTS, COMPOSE_STATES
from odcs.server.mbs import ModuleLookupError
from odcs.server.pungi import PungiSourceType
from odcs.server.backend import (
    resolve_compose,
    get_reusable_compose,
    generate_compose,
    generate_pulp_compose,
    generate_pungi_compose,
    validate_pungi_compose,
    koji_get_inherited_tags,
)
from odcs.server.utils import makedirs
import odcs.server.backend
from .utils import ModelsBaseTest, AnyStringWith

from .mbs import mock_mbs

thisdir = os.path.abspath(os.path.dirname(__file__))


class TestBackend(ModelsBaseTest):
    def test_resolve_compose_repo(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.REPO,
            os.path.join(thisdir, "repo"),
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()
        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.koji_event, 1496834159)

    @mock_mbs()
    def test_resolve_compose_module(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleA:f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(
            c.source,
            " ".join(
                [
                    "moduleA:f26:20170809000000:00000000",
                    "moduleB:f26:20170808000000:00000000",
                    "moduleC:f26:20170807000000:00000000",
                    "moduleD:f26:20170806000000:00000000",
                ]
            ),
        )

    @mock_mbs()
    def test_resolve_compose_module_empty_source(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.source, "")

    @mock_mbs()
    def test_resolve_compose_module_include_done_modules(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule:master",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["include_done_modules"],
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.source, "testmodule:master:20180515074419:00000000")

    @mock_mbs()
    def test_resolve_compose_module_include_done_modules_full_nsvc(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testmodule:master:20180515074419:00000000",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.source, "testmodule:master:20180515074419:00000000")

    @mock_mbs()
    def test_resolve_compose_module_devel(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleA:f26 moduleA-devel:f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(
            c.source,
            " ".join(
                [
                    "moduleA-devel:f26:20170809000000:00000000",
                    "moduleA:f26:20170809000000:00000000",
                    "moduleB:f26:20170808000000:00000000",
                    "moduleC:f26:20170807000000:00000000",
                    "moduleD:f26:20170806000000:00000000",
                ]
            ),
        )

    @mock_mbs()
    def test_resolve_compose_module_devel_deps_resolving(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleA-devel:f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(
            c.source,
            " ".join(
                [
                    "moduleA-devel:f26:20170809000000:00000000",
                    "moduleA:f26:20170809000000:00000000",
                    "moduleB:f26:20170808000000:00000000",
                    "moduleC:f26:20170807000000:00000000",
                    "moduleD:f26:20170806000000:00000000",
                ]
            ),
        )

    @mock_mbs()
    def test_resolve_compose_module_multiple_contexts_no_deps(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testcontexts:master:1",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["no_deps"],
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(
            c.source, " ".join(["testcontexts:master:1:a", "testcontexts:master:1:b"])
        )

    @mock_mbs()
    def test_resolve_compose_module_multiple_contexts_deps(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "testcontexts:master:1",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(
            c.source,
            " ".join(
                [
                    "parent:master:1:a",
                    "parent:master:1:b",
                    "testcontexts:master:1:a",
                    "testcontexts:master:1:b",
                ]
            ),
        )

    @mock_mbs()
    def test_resolve_compose_module_no_deps(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleA:f26 moduleA:f26",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["no_deps"],
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.source, "moduleA:f26:20170809000000:00000000")

    @mock_mbs()
    def expect_module_lookup_error(self, source, match, flags=0):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            source,
            COMPOSE_RESULTS["repository"],
            3600,
            flags=flags,
        )
        db.session.commit()

        with six.assertRaisesRegex(self, ModuleLookupError, match):
            resolve_compose(c)

    @mock_mbs(1)
    def test_resolve_compose_module_mmdv1(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleA:f26",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(
            c.source,
            " ".join(
                [
                    "moduleA:f26:20170809000000:00000000",
                    "moduleB:f26:20170808000000:00000000",
                    "moduleC:f26:20170807000000:00000000",
                    "moduleD:f26:20170806000000:00000000",
                ]
            ),
        )

    @mock_mbs(1)
    def test_resolve_compose_module_no_deps_mmdv1(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleA:f26 moduleA:f26",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["no_deps"],
        )
        db.session.commit()

        resolve_compose(c)
        db.session.commit()

        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.source, "moduleA:f26:20170809000000:00000000")

    @mock_mbs(1)
    def expect_module_lookup_error_mmdv1(self, source, match, flags=0):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            source,
            COMPOSE_RESULTS["repository"],
            3600,
            flags=flags,
        )
        db.session.commit()

        with six.assertRaisesRegex(self, ModuleLookupError, match):
            resolve_compose(c)

    def test_resolve_compose_module_not_found(self):
        self.expect_module_lookup_error("moduleA:f30", "Failed to find")

    def test_resolve_compose_module_not_found2(self):
        self.expect_module_lookup_error("moduleA:f26:00000000000000", "Failed to find")

    def test_resolve_compose_module_conflict(self):
        self.expect_module_lookup_error(
            "moduleA:f26:20170809000000 moduleA:f26:20170805000000", "conflicts with"
        )

    @mock_mbs()
    def test_resolve_compose_module_not_conflict(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleB:f26 moduleB:f27",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["no_deps"],
        )
        db.session.commit()

        resolve_compose(c)

    @mock_mbs(1)
    def test_resolve_compose_module_not_conflict_mmdv1(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "moduleB:f26 moduleB:f27",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["no_deps"],
        )
        db.session.commit()

        resolve_compose(c)

    def test_resolve_compose_module_dep_not_found(self):
        self.expect_module_lookup_error(
            "moduleB:f26 moduleB:f27",
            "Failed to find module moduleC:f27 in ready state in the MBS.",
        )

    @patch("odcs.server.backend.create_koji_session")
    def test_resolve_compose_repo_no_override_koji_event(self, create_koji_session):
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.getLastEvent.return_value = {"id": 123}

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        c.koji_event = 1
        db.session.commit()

        resolve_compose(c)
        db.session.commit()
        db.session.expire_all()
        c = db.session.query(Compose).filter(Compose.id == 1).one()
        self.assertEqual(c.koji_event, 1)

    def test_get_reusable_compose(self):
        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.REPO,
            os.path.join(thisdir, "repo"),
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        resolve_compose(old_c)
        old_c.state = COMPOSE_STATES["done"]
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.REPO,
            os.path.join(thisdir, "repo"),
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        resolve_compose(c)
        db.session.add(old_c)
        db.session.add(c)
        db.session.commit()

        reused_c = get_reusable_compose(c)
        self.assertEqual(reused_c, old_c)

    def test_koji_get_inherited_tags_unknown_tag(self):
        koji_session = MagicMock()
        koji_session.getTag.return_value = None

        with six.assertRaisesRegex(self, ValueError, "Unknown Koji tag foo."):
            koji_get_inherited_tags(koji_session, "foo")

    @patch("odcs.server.backend.koji_get_inherited_tags")
    @patch("odcs.server.backend.create_koji_session")
    def test_get_reusable_tag_compose(
        self, create_koji_session, koji_get_inherited_tags
    ):
        koji_get_inherited_tags.return_value = ["foo", "bar"]
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.tagChangedSinceEvent.return_value = False

        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        old_c.koji_event = 1
        old_c.state = COMPOSE_STATES["done"]
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        c.koji_event = 2
        db.session.add(old_c)
        db.session.add(c)
        db.session.commit()

        reused_c = get_reusable_compose(c)
        self.assertEqual(reused_c, old_c)

    @patch("odcs.server.backend.koji_get_inherited_tags")
    @patch("odcs.server.backend.create_koji_session")
    def test_get_reusable_tag_compose_none_koji_event(
        self, create_koji_session, koji_get_inherited_tags
    ):
        koji_get_inherited_tags.return_value = ["foo", "bar"]
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.tagChangedSinceEvent.return_value = False

        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        old_c.koji_event = None
        old_c.state = COMPOSE_STATES["done"]
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        c.koji_event = 2
        db.session.add(old_c)
        db.session.add(c)
        db.session.commit()

        reused_c = get_reusable_compose(c)
        self.assertEqual(reused_c, old_c)

    @patch("odcs.server.backend.koji_get_inherited_tags")
    @patch("odcs.server.backend.create_koji_session")
    def test_get_reusable_tag_compose_tag_changed(
        self, create_koji_session, koji_get_inherited_tags
    ):
        koji_get_inherited_tags.return_value = ["foo", "bar"]
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.tagChangedSinceEvent.return_value = True

        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        old_c.koji_event = 1
        old_c.state = COMPOSE_STATES["done"]
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        c.koji_event = 2
        db.session.add(old_c)
        db.session.add(c)
        db.session.commit()

        reused_c = get_reusable_compose(c)
        self.assertEqual(reused_c, None)

    @patch("odcs.server.backend.koji_get_inherited_tags")
    @patch("odcs.server.backend.create_koji_session")
    def test_get_reusable_tag_compose_renew(
        self, create_koji_session, koji_get_inherited_tags
    ):
        koji_get_inherited_tags.return_value = ["foo", "bar"]
        koji_session = MagicMock()
        create_koji_session.return_value = koji_session
        koji_session.tagChangedSinceEvent.return_value = False

        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        old_c.koji_event = 10
        old_c.state = COMPOSE_STATES["done"]
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo",
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
        )
        # c.koji_event is lower than old_c.koji_event, because "c" is actually
        # older than old_c and we are testing that its renewal does not reuse
        # the newer "c" compose.
        c.koji_event = 9
        db.session.add(old_c)
        db.session.add(c)
        db.session.commit()

        reused_c = get_reusable_compose(c)
        self.assertEqual(reused_c, None)

    def test_get_reusable_compose_attrs_not_the_same(self):
        old_c = Compose.create(
            db.session,
            "me",
            PungiSourceType.REPO,
            os.path.join(thisdir, "repo"),
            COMPOSE_RESULTS["repository"],
            3600,
            packages="ed",
            sigkeys="123",
            builds="foo-1-1",
        )
        old_c.state = COMPOSE_STATES["done"]
        resolve_compose(old_c)
        db.session.add(old_c)
        db.session.commit()

        attrs = {}
        attrs["packages"] = "ed foo"
        attrs["builds"] = "foo-1-1 bar-1-1"
        attrs["sigkeys"] = "321"
        attrs["koji_event"] = 123456
        attrs["source"] = "123"
        attrs["arches"] = "ppc64 x86_64"
        attrs["multilib_arches"] = "x86_64 i686"
        attrs["multilib_method"] = 1
        attrs["lookaside_repos"] = "foo bar"
        attrs["modular_koji_tags"] = "f26-modules"
        attrs["module_defaults_url"] = "git://localhost/x.git#branch"
        attrs["target_dir"] = "private"
        for attr, value in attrs.items():
            c = Compose.create(
                db.session,
                "me",
                PungiSourceType.REPO,
                os.path.join(thisdir, "repo"),
                COMPOSE_RESULTS["repository"],
                3600,
                packages="ed",
                sigkeys="123",
                builds="foo-1-1",
            )
            setattr(c, attr, value)

            # Do not resolve compose for non-existing source and in case we
            # change koji_event, because it would be overwriten.
            if attr not in ["source", "koji_event"]:
                resolve_compose(c)

            db.session.add(c)
            db.session.commit()
            reused_c = get_reusable_compose(c)
            self.assertEqual(reused_c, None)

    @patch("odcs.server.pulp.Pulp._rest_post")
    @patch("odcs.server.backend._write_repo_file")
    def test_generate_pulp_compose(self, _write_repo_file, pulp_rest_post):
        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
            {
                "notes": {
                    "relative_url": "content/2/x86_64/os",
                    "content_set": "foo-2",
                    "arch": "x86_64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                }
            },
            {
                "notes": {
                    "relative_url": "content/3/ppc64/os",
                    "content_set": "foo-3",
                    "arch": "ppc64",
                    "signatures": "SIG1,SIG3",
                    "product_versions": "",
                }
            },
        ]

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.PULP,
            "foo-1 foo-2 foo-3",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        with patch.object(
            odcs.server.backend.conf, "pulp_server_url", "https://localhost/"
        ):
            generate_pulp_compose(c)

        expected_query = {
            "criteria": {
                "fields": ["notes"],
                "filters": {
                    "notes.content_set": {"$in": ["foo-1", "foo-2", "foo-3"]},
                    "notes.include_in_download_service": "True",
                },
            }
        }
        pulp_rest_post.assert_called_once_with("repositories/search/", expected_query)

        expected_repofile = """
[foo-1]
name=foo-1
baseurl=http://localhost/content/1/x86_64/os
enabled=1
gpgcheck=0

[foo-2]
name=foo-2
baseurl=http://localhost/content/2/x86_64/os
enabled=1
gpgcheck=0

[foo-3]
name=foo-3
baseurl=http://localhost/content/3/ppc64/os
enabled=1
gpgcheck=0
"""
        _write_repo_file.assert_called_once_with(c, expected_repofile)

        self.assertEqual(c.state, COMPOSE_STATES["done"])
        self.assertEqual(c.state_reason, "Compose is generated successfully")
        self.assertEqual(len(c.arches.split(" ")), 2)
        self.assertEqual(set(c.arches.split(" ")), set(["x86_64", "ppc64"]))
        self.assertEqual(len(c.sigkeys.split(" ")), 3)
        self.assertEqual(set(c.sigkeys.split(" ")), set(["SIG1", "SIG2", "SIG3"]))

    @patch("odcs.server.pulp.Pulp._rest_post")
    @patch("odcs.server.backend._write_repo_file")
    @patch("os.symlink")
    def test_generate_pulp_compose_include_inpublished_pulp_repos_passed(
        self, symlink, _write_repo_file, pulp_rest_post
    ):
        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "ppc64",
                    "signatures": "SIG1,SIG2",
                },
            },
        ]

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.PULP,
            "foo-1 foo-2",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["include_unpublished_pulp_repos"],
        )
        db.session.add(c)
        db.session.commit()

        with patch.object(
            odcs.server.backend.conf, "pulp_server_url", "https://localhost/"
        ):
            generate_compose(1)

        expected_query = {
            "criteria": {
                "fields": ["notes"],
                "filters": {"notes.content_set": {"$in": ["foo-1", "foo-2"]}},
            }
        }
        pulp_rest_post.assert_called_once_with("repositories/search/", expected_query)
        symlink.assert_not_called()

    @patch("odcs.server.pulp.Pulp._rest_post")
    @patch("odcs.server.backend._write_repo_file")
    def test_generate_pulp_compose_content_set_not_found(
        self, _write_repo_file, pulp_rest_post
    ):
        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "ppc64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
        ]

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.PULP,
            "foo-1 foo-2",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        with patch.object(
            odcs.server.backend.conf, "pulp_server_url", "https://localhost/"
        ):
            generate_compose(1)

        expected_query = {
            "criteria": {
                "fields": ["notes"],
                "filters": {
                    "notes.content_set": {"$in": ["foo-1", "foo-2"]},
                    "notes.include_in_download_service": "True",
                },
            }
        }
        pulp_rest_post.assert_called_once_with("repositories/search/", expected_query)
        _write_repo_file.assert_not_called()

        c1 = Compose.query.filter(Compose.id == 1).one()
        self.assertEqual(c1.state, COMPOSE_STATES["failed"])
        six.assertRegex(
            self,
            c1.state_reason,
            r"Error while generating compose: Failed to find all the content_sets.*",
        )

    @patch("odcs.server.pulp.Pulp._rest_post")
    @patch("odcs.server.backend._write_repo_file")
    def test_generate_pulp_compose_content_set_not_found_allow_absent(
        self, _write_repo_file, pulp_rest_post
    ):
        pulp_rest_post.return_value = [
            {
                "notes": {
                    "relative_url": "content/1/x86_64/os",
                    "content_set": "foo-1",
                    "arch": "ppc64",
                    "signatures": "SIG1,SIG2",
                    "product_versions": "",
                },
            },
        ]

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.PULP,
            "foo-1 foo-2",
            COMPOSE_RESULTS["repository"],
            3600,
            flags=COMPOSE_FLAGS["ignore_absent_pulp_repos"],
        )
        db.session.add(c)
        db.session.commit()

        with patch.object(
            odcs.server.backend.conf, "pulp_server_url", "https://localhost/"
        ):
            generate_compose(1)

        expected_query = {
            "criteria": {
                "fields": ["notes"],
                "filters": {
                    "notes.content_set": {"$in": ["foo-1", "foo-2"]},
                    "notes.include_in_download_service": "True",
                },
            }
        }
        pulp_rest_post.assert_called_once_with("repositories/search/", expected_query)
        expected_repofile = """
[foo-1]
name=foo-1
baseurl=http://localhost/content/1/x86_64/os
enabled=1
gpgcheck=0
"""
        _write_repo_file.assert_called_once_with(c, expected_repofile)

        c1 = Compose.query.filter(Compose.id == 1).one()
        self.assertEqual(c1.state, COMPOSE_STATES["done"])
        self.assertEqual(c1.source, "foo-1")

    @patch("odcs.server.backend.resolve_compose")
    @patch("odcs.server.backend.generate_pungi_compose")
    @patch("odcs.server.pungi.PungiLogs.get_error_string")
    def test_generate_compose_exception(
        self, get_error_string, generate_pungi_compose, resolve_compose
    ):
        get_error_string.return_value = "Compose failed for unknown reason."
        generate_pungi_compose.side_effect = RuntimeError(
            "Expected exception, see %s" % os.path.join(conf.target_dir, "foo.log")
        )

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo-1",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        generate_compose(1)

        get_error_string.assert_called_once()

        c1 = Compose.query.filter(Compose.id == 1).one()
        self.assertEqual(c1.state, COMPOSE_STATES["failed"])
        six.assertRegex(
            self,
            c1.state_reason,
            r"Error while generating compose: Expected exception, see "
            "http://localhost/odcs/foo.log\n"
            "Compose failed for unknown reason*",
        )

    @patch("odcs.server.backend.resolve_compose")
    @patch("odcs.server.backend.generate_pungi_compose")
    @patch("odcs.server.pungi.PungiLogs.get_error_string")
    def test_generate_compose_pungi_logs_exceptions(
        self, get_error_string, generate_pungi_compose, resolve_compose
    ):
        get_error_string.side_effect = RuntimeError("PungiLogs Expected exception")
        generate_pungi_compose.side_effect = RuntimeError("Expected exception")

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo-1",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        generate_compose(1)

        c1 = Compose.query.filter(Compose.id == 1).one()
        self.assertEqual(c1.state, COMPOSE_STATES["failed"])
        six.assertRegex(
            self,
            c1.state_reason,
            r"Error while generating compose: Expected exception*",
        )

    @patch("odcs.server.backend.tag_changed", return_value=True)
    @patch("odcs.server.backend.create_koji_session")
    def test_resolve_compose_from_koji_tag_get_last_event_if_tag_changed(
        self, create_koji_session, tag_changed
    ):
        session = create_koji_session.return_value
        fake_koji_event = {"id": 234567}
        session.getLastEvent.return_value = fake_koji_event

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo-1",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        with patch.dict("odcs.server.backend.LAST_EVENTS_CACHE", {"foo-1": 123456}):
            resolve_compose(c)
            c.koji_event = fake_koji_event["id"]

    @patch("odcs.server.backend.tag_changed")
    @patch("odcs.server.backend.create_koji_session")
    def test_resolve_compose_from_koji_tag_reuse_koji_event_if_tag_not_changed(
        self, create_koji_session, tag_changed
    ):
        tag_changed.return_value = False
        session = create_koji_session.return_value

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo-1",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        with patch.dict("odcs.server.backend.LAST_EVENTS_CACHE", {"foo-1": 123456}):
            resolve_compose(c)
            c.koji_event = 123456

            session.getLastEvent.assert_not_called()

    @patch("odcs.server.backend.tag_changed")
    @patch("odcs.server.backend.create_koji_session")
    def test_resolve_compose_from_koji_tag_get_last_koji_event_if_tag_not_cached(
        self, create_koji_session, tag_changed
    ):
        fake_koji_event = {"id": 789065}
        session = create_koji_session.return_value
        session.getLastEvent.return_value = fake_koji_event

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "foo-1",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        with patch.dict("odcs.server.backend.LAST_EVENTS_CACHE", {"bar-2": 123456}):
            resolve_compose(c)
            c.koji_event = fake_koji_event["id"]

            tag_changed.assert_not_called()

    @patch("odcs.server.mbs.MBS.validate_module_list")
    @patch("odcs.server.mbs.MBS.get_latest_modules")
    def test_resolve_compose_module_filter_base_module(
        self, get_latest_modules, validate_module_list
    ):
        modules = [
            {"name": "foo", "stream": "0", "version": 1, "context": "x"},
            {"name": "bar", "stream": "0", "version": 1, "context": "y"},
        ]
        get_latest_modules.return_value = modules
        validate_module_list.return_value = modules + [
            {"name": "platform", "stream": "0", "version": 1, "context": "z"}
        ]

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.MODULE,
            "foo:0 bar:0",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        # Default conf.base_module_names includes "platform" module, so it
        # should be removed from the source list.
        resolve_compose(c)
        self.assertEqual(c.source, "bar:0:1:y foo:0:1:x")

        # Now try removing "platform" from the conf.base_module_names, so it
        # should appear in a compose source.
        with patch.object(
            odcs.server.config.Config, "base_module_names", new=["random_name"]
        ):
            resolve_compose(c)
            self.assertEqual(c.source, "bar:0:1:y foo:0:1:x platform:0:1:z")

    @patch("odcs.server.pungi_compose.PungiCompose.get_rpms_data")
    def test_resolve_compose_pungi_compose_source_type(self, get_rpms_data):
        get_rpms_data.return_value = {
            "sigkeys": set(["sigkey1", None]),
            "arches": set(["x86_64"]),
            "builds": {
                "flatpak-rpm-macros-29-6.module+125+c4f5c7f2": set(
                    [
                        "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.src",
                        "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.x86_64",
                    ]
                ),
                "flatpak-runtime-config-29-4.module+125+c4f5c7f2": set(
                    [
                        "flatpak-runtime-config-0:29-4.module+125+c4f5c7f2.src",
                        "flatpak-runtime-config2-0:29-4.module+125+c4f5c7f2.x86_64",
                    ]
                ),
            },
        }

        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.PUNGI_COMPOSE,
            "http://localhost/compose/Temporary",
            COMPOSE_RESULTS["repository"],
            3600,
        )
        db.session.add(c)
        db.session.commit()

        resolve_compose(c)
        self.assertEqual(c.sigkeys.split(" "), ["sigkey1", ""])
        self.assertEqual(c.arches.split(" "), ["x86_64"])
        self.assertEqual(
            set(c.builds.split(" ")),
            set(
                [
                    "flatpak-rpm-macros-29-6.module+125+c4f5c7f2",
                    "flatpak-runtime-config-29-4.module+125+c4f5c7f2",
                ]
            ),
        )
        self.assertEqual(
            set(c.packages.split(" ")),
            set(
                [
                    "flatpak-rpm-macros",
                    "flatpak-runtime-config",
                    "flatpak-runtime-config2",
                ]
            ),
        )


class TestGeneratePungiCompose(ModelsBaseTest):
    def setUp(self):
        super(TestGeneratePungiCompose, self).setUp()

        self.patch_resolve_compose = patch("odcs.server.backend.resolve_compose")
        self.resolve_compose = self.patch_resolve_compose.start()

        self.patch_get_reusable_compose = patch(
            "odcs.server.backend.get_reusable_compose"
        )
        self.get_reusable_compose = self.patch_get_reusable_compose.start()
        self.get_reusable_compose.return_value = False

        self.patch_write_repo_file = patch("odcs.server.backend._write_repo_file")
        self.write_repo_file = self.patch_write_repo_file.start()

        self.patch_is_cached = patch("odcs.server.backend.KojiTagCache.is_cached")
        self.is_cached = self.patch_is_cached.start()

        self.patch_reuse_cached = patch("odcs.server.backend.KojiTagCache.reuse_cached")
        self.reuse_cached = self.patch_reuse_cached.start()

        self.patch_update_cache = patch("odcs.server.backend.KojiTagCache.update_cache")
        self.update_cache = self.patch_update_cache.start()

        self.patch_validate_pungi_compose = patch(
            "odcs.server.backend.validate_pungi_compose"
        )
        self.validate_pungi_compose = self.patch_validate_pungi_compose.start()

        # Mocked method to store Pungi.pungi_cfg to self.pungi_cfg, so we can
        # run asserts against it.
        self.pungi_config = None
        self.old_compose = None

        def fake_pungi_run(pungi_cls, compose):
            self.pungi_config = pungi_cls.pungi_cfg
            self.old_compose = pungi_cls.old_compose

        self.patch_pungi_run = patch("odcs.server.pungi.Pungi.run", autospec=True)
        self.pungi_run = self.patch_pungi_run.start()
        self.pungi_run.side_effect = fake_pungi_run

    def tearDown(self):
        super(TestGeneratePungiCompose, self).tearDown()
        self.patch_resolve_compose.stop()
        self.patch_get_reusable_compose.stop()
        self.patch_write_repo_file.stop()
        self.patch_pungi_run.stop()
        self.patch_reuse_cached.stop()
        self.patch_update_cache.stop()
        self.validate_pungi_compose.stop()
        self.patch_is_cached.stop()
        self.pungi_config = None

    def test_generate_pungi_compose(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1 pkg2 pkg3",
            arches="x86_64 s390",
            multilib_arches="i686 x86_64",
            multilib_method=1,
        )
        c.id = 1

        generate_pungi_compose(c)

        self.resolve_compose.assert_called_once_with(c)
        self.get_reusable_compose.assert_called_once_with(c)
        self.write_repo_file.assert_called_once_with(c)
        self.is_cached.assert_called_once_with(c)
        self.reuse_cached.assert_called_once_with(c)
        self.update_cache.assert_called_once_with(c)
        self.validate_pungi_compose.assert_called_once_with(c)

        self.assertEqual(self.pungi_config.gather_method, "deps")
        self.assertEqual(self.pungi_config.pkgset_koji_inherit, True)
        self.assertEqual(set(self.pungi_config.arches), set(["x86_64", "s390"]))
        self.assertEqual(
            set(self.pungi_config.multilib_arches), set(["i686", "x86_64"])
        )
        self.assertEqual(self.pungi_config.multilib_method, ["runtime"])

    def test_generate_pungi_compose_multiarch_arches_None(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1 pkg2 pkg3",
            arches="x86_64 s390",
            multilib_arches=None,
            multilib_method=None,
        )
        c.id = 1

        generate_pungi_compose(c)
        self.assertEqual(set(self.pungi_config.multilib_arches), set([]))
        self.assertEqual(self.pungi_config.multilib_method, [])

    def test_generate_pungi_compose_nodeps(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1 pkg2 pkg3",
            flags=COMPOSE_FLAGS["no_deps"],
        )
        c.id = 1

        generate_pungi_compose(c)
        self.assertEqual(self.pungi_config.gather_method, "nodeps")
        self.assertEqual(self.pungi_config.pkgset_koji_inherit, True)

    def test_generate_pungi_compose_noinheritance(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1 pkg2 pkg3",
            flags=COMPOSE_FLAGS["no_inheritance"],
        )
        c.id = 1

        generate_pungi_compose(c)
        self.assertEqual(self.pungi_config.gather_method, "deps")
        self.assertEqual(self.pungi_config.pkgset_koji_inherit, False)

    def test_generate_pungi_compose_builds(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            builds="foo-1-1 bar-1-1",
            flags=COMPOSE_FLAGS["no_inheritance"],
        )
        c.id = 1

        generate_pungi_compose(c)
        self.assertEqual(self.pungi_config.builds, ["foo-1-1", "bar-1-1"])

    def test_generate_pungi_compose_source_type_build(self):
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.BUILD,
            "x",
            COMPOSE_RESULTS["repository"],
            60,
            builds="foo-1-1 bar-1-1",
            flags=COMPOSE_FLAGS["no_inheritance"],
        )
        c.id = 1

        generate_pungi_compose(c)
        self.assertEqual(self.pungi_config.koji_tag, None)
        self.assertEqual(self.pungi_config.builds, ["foo-1-1", "bar-1-1"])

    @patch.object(
        odcs.server.config.Config,
        "raw_config_urls",
        new={
            "pungi_cfg": {
                "url": "git://localhost/test.git",
                "config_filename": "pungi.conf",
            }
        },
    )
    @patch("odcs.server.utils.makedirs")
    @patch("os.symlink")
    @patch("os.unlink")
    @patch("odcs.server.pungi.PungiLogs.get_config_dump")
    @patch("odcs.server.backend._write_repo_file")
    def test_generate_pungi_compose_raw_config(
        self, write_repo_file, config_dump, unlink, symlink, makedirs
    ):
        self.is_cached.return_value = False
        config_dump.return_value = "fake\npungi\nconf\n"
        c = Compose.create(
            db.session,
            "me",
            PungiSourceType.RAW_CONFIG,
            "pungi_cfg#hash",
            COMPOSE_RESULTS["repository"],
            60,
        )
        c.compose_type = "production"
        c.pungi_compose_id = "compose-1-10-2020110.n.0"
        c.id = 1

        fake_raw_config_urls = {
            "pungi_cfg": {
                "url": "git://localhost/test.git",
                "config_filename": "pungi.conf",
            }
        }
        with patch.object(conf, "raw_config_urls", new=fake_raw_config_urls):
            generate_pungi_compose(c)

        self.assertEqual(c.pungi_config_dump, "fake\npungi\nconf\n")
        self.assertEqual(
            self.pungi_config.pungi_cfg,
            {
                "url": "git://localhost/test.git",
                "config_filename": "pungi.conf",
                "commit": "hash",
            },
        )
        self.assertEqual(self.old_compose, AnyStringWith("/test_composes/production"))

        makedirs.assert_called_once_with(AnyStringWith("/test_composes/production"))
        symlink.assert_has_calls(
            [
                call(
                    "../odcs-1",
                    AnyStringWith("/test_composes/production/compose-1-10-2020110.n.0"),
                ),
                call(
                    "../odcs-1",
                    AnyStringWith("/test_composes/production/latest-compose-1"),
                ),
            ]
        )
        unlink.assert_called_with(
            AnyStringWith("/test_composes/production/latest-compose-1")
        )
        write_repo_file.assert_not_called()


class TestValidatePungiCompose(ModelsBaseTest):
    """Test validate_pungi_compose"""

    def setUp(self):
        super(TestValidatePungiCompose, self).setUp()

        self.c = Compose.create(
            db.session,
            "me",
            PungiSourceType.KOJI_TAG,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            packages="pkg1 pkg2 pkg3",
        )
        db.session.commit()

        # Remove any previous toplevel_dir.
        if os.path.exists(self.c.toplevel_dir):
            shutil.rmtree(self.c.toplevel_dir)

        compose_dir = os.path.join(self.c.toplevel_dir, "compose")
        metadata_dir = os.path.join(compose_dir, "metadata")
        self.rpms_metadata = os.path.join(metadata_dir, "rpms.json")
        makedirs(metadata_dir)

        rm = Rpms()
        rm.header.version = "1.0"
        rm.compose.id = "Me-26-20161212.0"
        rm.compose.type = "production"
        rm.compose.date = "20161212"
        rm.compose.respin = 0

        # pkg1
        rm.add(
            "Temporary",
            "x86_64",
            "pkg1-0:2.18-11.fc26.x86_64.rpm",
            path="Temporary/x86_64/os/Packages/p/pkg1-2.18-11.fc26.x86_64.rpm",
            sigkey="246110c1",
            category="binary",
            srpm_nevra="pkg1-0:2.18-11.fc26.src.rpm",
        )
        rm.add(
            "Temporary",
            "x86_64",
            "pkg1-lib-0:2.18-11.fc26.x86_64.rpm",
            path="Temporary/x86_64/os/Packages/p/pkg1-lib-2.18-11.fc26.x86_64.rpm",
            sigkey="246110c1",
            category="binary",
            srpm_nevra="pkg1-0:2.18-11.fc26.src.rpm",
        )
        rm.add(
            "Temporary",
            "x86_64",
            "pkg1-0:2.18-11.fc26.src.rpm",
            path="Temporary/source/SRPMS/p/pkg1-2.18-11.fc26.x86_64.rpm",
            sigkey="246110c1",
            category="source",
        )
        # pkg2
        rm.add(
            "Temporary",
            "x86_64",
            "pkg2-0:2.18-11.fc26.x86_64.rpm",
            path="Temporary/x86_64/os/Packages/p/pkg2-0.18-11.fc26.x86_64.rpm",
            sigkey="246110c1",
            category="binary",
            srpm_nevra="pkg2-0:0.18-11.fc26.src.rpm",
        )
        rm.add(
            "Temporary",
            "x86_64",
            "pkg2-lib-0:2.18-11.fc26.x86_64.rpm",
            path="Temporary/x86_64/os/Packages/p/pkg2-lib-0.18-11.fc26.x86_64.rpm",
            sigkey="246110c1",
            category="binary",
            srpm_nevra="pkg2-0:0.18-11.fc26.src.rpm",
        )
        rm.add(
            "Temporary",
            "x86_64",
            "pkg2-0:0.18-11.fc26.src.rpm",
            path="Temporary/source/SRPMS/p/pkg2-0.18-11.fc26.x86_64.rpm",
            sigkey="246110c1",
            category="source",
        )
        rm.dump(self.rpms_metadata)

    def tearDown(self):
        shutil.rmtree(self.c.toplevel_dir)
        super(TestValidatePungiCompose, self).tearDown()

    def test_missing_packages(self):
        with six.assertRaisesRegex(self, RuntimeError, "not present.+pkg3"):
            validate_pungi_compose(self.c)

    def test_all_packages_are_included(self):
        self.c.packages = "pkg1 pkg1-lib pkg2 pkg2-lib"
        db.session.commit()

        validate_pungi_compose(self.c)



from mock import patch, Mock, call
from datetime import datetime, timedelta

import pytest
from .utils import ModelsBaseTest

from odcs.server import conf, db
from odcs.server.celery_tasks import TaskRouter, reschedule_waiting_composes
from odcs.common.types import COMPOSE_STATES, COMPOSE_RESULTS
from odcs.server.pungi import PungiSourceType
from odcs.server.models import Compose


class TestCeleryRouter:
    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_empty_rule(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {"source_type": 3}

        mock_conf = {
            "routing_rules": {
                "odcs.server.celery_tasks.generate_pungi_compose": {
                    "pungi_composes": {},
                    "other_composes": {"source_type": 4},
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        queue = tr.route_for_task(
            "odcs.server.celery_tasks.generate_pungi_compose", *args, **kwargs
        )
        assert queue == {"queue": "pungi_composes"}

    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_default_queue(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {"source_type": 3}

        mock_conf = {
            "routing_rules": {
                "some.other.task": {
                    "pungi_composes": {},
                    "other_composes": {"source_type": 4},
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        queue = tr.route_for_task(
            "odcs.server.celery_tasks.generate_pungi_compose", *args, **kwargs
        )
        assert queue == {"queue": "default_queue"}

    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_rule_with_single_property(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {"source_type": 3}

        mock_conf = {
            "routing_rules": {
                "odcs.server.celery_tasks.generate_pungi_compose": {
                    "pungi_composes": {"source_type": 3},
                    "other_composes": {"source_type": 4},
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        queue = tr.route_for_task(
            "odcs.server.celery_tasks.generate_pungi_compose", *args, **kwargs
        )
        assert queue == {"queue": "pungi_composes"}

    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_rule_with_list_property(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {
            "source_type": 3,
            "user": "mprahl",
        }

        mock_conf = {
            "routing_rules": {
                "odcs.server.celery_tasks.generate_pungi_compose": {
                    "pungi_composes": {
                        "source_type": 3,
                        "user": ["mcurlej", "jkaluza"],
                    },
                    "other_composes": {
                        "source_type": 3,
                        "user": ["mprahl", "lucarval"],
                    },
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        queue = tr.route_for_task(
            "odcs.server.celery_tasks.generate_pungi_compose", *args, **kwargs
        )
        assert queue == {"queue": "other_composes"}

    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_cleanup_queue(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {"source_type": 3}

        mock_conf = {
            "routing_rules": {
                "odcs.server.celery_tasks.generate_pungi_compose": {
                    "pungi_composes": {"source_type": 3},
                    "other_composes": {"source_type": 4},
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        queue = tr.route_for_task(
            "odcs.server.celery_tasks.run_cleanup", *args, **kwargs
        )
        assert queue == {"queue": conf.celery_cleanup_queue}

    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_invalid_rule_property_exception(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {"source_type": 3}

        mock_conf = {
            "routing_rules": {
                "odcs.server.celery_tasks.generate_pungi_compose": {
                    "pungi_composes": {"bad_compose_prop": 3},
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        with pytest.raises(ValueError) as e:
            tr.route_for_task(
                "odcs.server.celery_tasks.generate_pungi_compose", *args, **kwargs
            )
            assert "invalid property" in e.args[0]
            assert "bad_compose_prop" in e.args[0]

    @patch("odcs.server.celery_tasks.get_odcs_compose")
    def test_rule_with_regexp(self, mock_get_compose):
        mock_compose = Mock()

        compose_md = {
            "source_type": 3,
            "source": "fedora30#commithash",
        }

        mock_conf = {
            "routing_rules": {
                "odcs.server.celery_tasks.generate_pungi_compose": {
                    "pungi_composes": {"source_type": 3, "source": "^fedora30#.*"},
                },
            },
            "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
            "default_queue": "default_queue",
        }

        tr = TaskRouter()
        tr.config = mock_conf

        mock_compose.json.return_value = compose_md
        mock_get_compose.return_value = mock_compose
        args = [[1], {}]
        kwargs = {}
        queue = tr.route_for_task(
            "odcs.server.celery_tasks.generate_pungi_compose", *args, **kwargs
        )
        assert queue == {"queue": "pungi_composes"}


class TestRescheduleWaitingComposes(ModelsBaseTest):
    def _add_test_compose(
        self,
        state,
        time_submitted=None,
        time_started=None,
        source_type=PungiSourceType.KOJI_TAG,
    ):
        compose = Compose.create(
            db.session,
            "unknown",
            source_type,
            "f26",
            COMPOSE_RESULTS["repository"],
            60,
            "",
            0,
        )
        compose.state = state
        compose.celery_task_id = "1"
        if time_submitted:
            compose.time_submitted = time_submitted
        if time_started:
            compose.time_started = time_started
        db.session.add(compose)
        db.session.commit()
        return compose

    @patch("odcs.server.celery_tasks.get_current_celery_task_ids")
    @patch("odcs.server.celery_tasks.schedule_compose")
    def test_reschedule_waiting_composes_generating_state(
        self, schedule_compose, task_ids
    ):
        task_ids.return_value = set(["2"])
        time_submitted = datetime.utcnow() - timedelta(minutes=5)
        composes = []
        for i in range(10):
            composes.append(
                self._add_test_compose(
                    COMPOSE_STATES["wait"], time_submitted=time_submitted
                )
            )
        composes = sorted(composes, key=lambda c: c.id)
        reschedule_waiting_composes()
        schedule_compose.assert_has_calls(
            [call(composes[0]), call(composes[1]), call(composes[2]), call(composes[3])]
        )

    @patch("odcs.server.celery_tasks.get_current_celery_task_ids")
    @patch("odcs.server.celery_tasks.schedule_compose")
    def test_reschedule_waiting_composes_generating_state_not_old_enough(
        self, schedule_compose, task_ids
    ):
        task_ids.return_value = set(["2"])
        composes = []
        for i in range(10):
            composes.append(self._add_test_compose(COMPOSE_STATES["wait"]))
        composes = sorted(composes, key=lambda c: c.id)
        reschedule_waiting_composes()
        schedule_compose.assert_not_called()

    @patch("odcs.server.celery_tasks.get_current_celery_task_ids")
    @patch("odcs.server.celery_tasks.schedule_compose")
    def test_reschedule_waiting_composes_generating_state_old(
        self, schedule_compose, task_ids
    ):
        task_ids.return_value = set(["2"])
        time_submitted = datetime.utcnow() - timedelta(days=5)
        composes = []
        for i in range(10):
            composes.append(
                self._add_test_compose(
                    COMPOSE_STATES["wait"], time_submitted=time_submitted
                )
            )
        composes = sorted(composes, key=lambda c: c.id)
        reschedule_waiting_composes()
        schedule_compose.assert_not_called()

    @patch("odcs.server.celery_tasks.get_current_celery_task_ids")
    @patch("odcs.server.celery_tasks.schedule_compose")
    def test_generate_lost_composes_generating_state(self, schedule_compose, task_ids):
        task_ids.return_value = set(["2"])
        composes = []
        for i in range(10):
            composes.append(self._add_test_compose(COMPOSE_STATES["generating"]))
        composes = sorted(composes, key=lambda c: c.id)
        reschedule_waiting_composes()
        schedule_compose.assert_not_called()



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from flask_script import Manager
from functools import wraps
import flask_migrate
import logging
import os
import ssl

from odcs.server import app, conf, db
from odcs.server import models


manager = Manager(app)
help_args = ("-?", "--help")
manager.help_args = help_args
migrations_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "migrations")
migrate = flask_migrate.Migrate(app, db, directory=migrations_dir)
manager.add_command("db", flask_migrate.MigrateCommand)


def console_script_help(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        import sys

        if any([arg in help_args for arg in sys.argv[1:]]):
            command = os.path.basename(sys.argv[0])
            print(
                """{0}

Usage: {0} [{1}]

See also:
  odcs-manager(1)""".format(
                    command, "|".join(help_args)
                )
            )
            sys.exit(2)
        r = f(*args, **kwargs)
        return r

    return wrapped


def _establish_ssl_context():
    if not conf.ssl_enabled:
        return None
    # First, do some validation of the configuration
    attributes = (
        "ssl_certificate_file",
        "ssl_certificate_key_file",
        "ssl_ca_certificate_file",
    )

    for attribute in attributes:
        value = getattr(conf, attribute, None)
        if not value:
            raise ValueError("%r could not be found" % attribute)
        if not os.path.exists(value):
            raise OSError("%s: %s file not found." % (attribute, value))

    # Then, establish the ssl context and return it
    ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    ssl_ctx.load_cert_chain(conf.ssl_certificate_file, conf.ssl_certificate_key_file)
    ssl_ctx.verify_mode = ssl.CERT_OPTIONAL
    ssl_ctx.load_verify_locations(cafile=conf.ssl_ca_certificate_file)
    return ssl_ctx


@console_script_help
@manager.command
def upgradedb():
    """Upgrades the database schema to the latest revision"""
    app.config["SERVER_NAME"] = "localhost"
    migrations_dir = os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "migrations"
    )
    with app.app_context():
        flask_migrate.upgrade(directory=migrations_dir)


@console_script_help
@manager.command
def cleardb():
    """Clears the database"""
    models.Event.query.delete()
    models.ArtifactBuild.query.delete()
    db.session.commit()


@manager.command
@console_script_help
def generatelocalhostcert():
    """Creates a public/private key pair for message signing and the frontend"""
    from OpenSSL import crypto

    cert_key = crypto.PKey()
    cert_key.generate_key(crypto.TYPE_RSA, 2048)

    with open(conf.ssl_certificate_key_file, "w") as cert_key_file:
        os.chmod(conf.ssl_certificate_key_file, 0o600)
        cert_key_file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, cert_key))

    cert = crypto.X509()
    msg_cert_subject = cert.get_subject()
    msg_cert_subject.C = "US"
    msg_cert_subject.ST = "MA"
    msg_cert_subject.L = "Boston"
    msg_cert_subject.O = "Development"  # noqa
    msg_cert_subject.CN = "localhost"
    cert.set_serial_number(2)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(315360000)  # 10 years
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(cert_key)
    cert_extensions = [
        crypto.X509Extension(
            "keyUsage", True, "digitalSignature, keyEncipherment, nonRepudiation"
        ),
        crypto.X509Extension("extendedKeyUsage", True, "serverAuth"),
    ]
    cert.add_extensions(cert_extensions)
    cert.sign(cert_key, "sha256")

    with open(conf.ssl_certificate_file, "w") as cert_file:
        cert_file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))


@console_script_help
@manager.command
def runssl(host=conf.host, port=conf.port, debug=conf.debug):
    """Runs the Flask app with the HTTPS settings configured in config.py"""
    logging.info("Starting ODCS frontend")

    ssl_ctx = _establish_ssl_context()
    app.run(host=host, port=port, ssl_context=ssl_ctx, debug=debug)


def manager_wrapper():
    """
    Runs the manager. We have separate method for this so we can use it in
    `console_scripts` part of setup.py
    """
    manager.run()


if __name__ == "__main__":
    manager_wrapper()



"""Add scratch_build_tasks field.

Revision ID: ca08065687c4
Revises: 7b143656694f
Create Date: 2020-09-03 11:20:54.035095

"""

# revision identifiers, used by Alembic.
revision = "ca08065687c4"
down_revision = "7b143656694f"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(
        "composes", sa.Column("scratch_build_tasks", sa.String(), nullable=True)
    )


def downgrade():
    op.drop_column("composes", "scratch_build_tasks")



# -*- coding: utf-8 -*-

# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import imp
import os
import sys

from odcs.server import logger
from odcs.common.types import COMPOSE_RESULTS, COMPOSE_FLAGS


def init_config(app):
    """
    Configure ODCS
    """
    config_module = None
    config_file = "/etc/odcs/config.py"
    config_section = "DevConfiguration"

    # automagically detect production environment:
    #   - existing and readable config_file presets ProdConfiguration
    try:
        with open(config_file):
            config_section = "ProdConfiguration"
    except (OSError, IOError) as e:
        # Use stderr here, because logging is not initialized so far...
        sys.stderr.write("WARN: Cannot open %s: %s\n" % (config_file, e.strerror))
        sys.stderr.write("WARN: DevConfiguration will be used.\n")

    # try getting config_file from os.environ
    if "ODCS_CONFIG_FILE" in os.environ:
        config_file = os.environ["ODCS_CONFIG_FILE"]
    # try getting config_section from os.environ
    if "ODCS_CONFIG_SECTION" in os.environ:
        config_section = os.environ["ODCS_CONFIG_SECTION"]
    # TestConfiguration shall only be used for running tests, otherwise...
    if any(
        [
            "nosetests" in arg
            or "noserunner.py" in arg
            or "py.test" in arg
            or "pytest.py" in arg
            for arg in sys.argv
        ]
    ):
        config_section = "TestConfiguration"
        from conf import config

        config_module = config
    # ...ODCS_DEVELOPER_ENV has always the last word
    # and overrides anything previously set before!
    # In any of the following cases, use configuration directly from ODCS
    # package -> /conf/config.py.

    elif "ODCS_DEVELOPER_ENV" in os.environ and os.environ[
        "ODCS_DEVELOPER_ENV"
    ].lower() in ("1", "on", "true", "y", "yes"):
        config_section = "DevConfiguration"
        from conf import config

        config_module = config
    # try loading configuration from file
    if not config_module:
        try:
            config_module = imp.load_source("odcs_runtime_config", config_file)
        except Exception:
            raise SystemError(
                "Configuration file {} was not found.".format(config_file)
            )

    # finally configure ODCS
    config_section_obj = getattr(config_module, config_section)
    conf = Config(config_section_obj)
    app.config.from_object(config_section_obj)
    return conf


class Config(object):
    """Class representing the odcs configuration."""

    _defaults = {
        "debug": {"type": bool, "default": False, "desc": "Debug mode"},
        "log_backend": {"type": str, "default": None, "desc": "Log backend"},
        "log_file": {"type": str, "default": "", "desc": "Path to log file"},
        "log_level": {"type": str, "default": 0, "desc": "Log level"},
        "net_timeout": {
            "type": int,
            "default": 120,
            "desc": "Global network timeout for read/write operations, in seconds.",
        },
        "net_retry_interval": {
            "type": int,
            "default": 30,
            "desc": "Global network retry interval for read/write operations, in seconds.",
        },
        "arches": {
            "type": list,
            "default": ["x86_64"],
            "desc": "Compose architectures.",
        },
        "pungi_koji": {
            "type": str,
            "default": "pungi-koji",
            "desc": "Name or full-path to pungi-koji binary.",
        },
        "pungi_config_validate": {
            "type": str,
            "default": "",
            "desc": "Name or full-path to pungi-config-validate binary. "
            "If set to empty string, no validation is done.",
        },
        "pungi_timeout": {
            "type": int,
            "default": 3600,
            "desc": "Time in seconds after which the local pungi-koji is "
            "killed and compose is marked as failed",
        },
        "mergerepo_timeout": {
            "type": int,
            "default": 1800,
            "desc": "Time in seconds after which the mergerepo_c is "
            "killed and compose is marked as failed",
        },
        "pungi_conf_path": {
            "type": str,
            "default": "/etc/odcs/pungi.conf",
            "desc": "Full path to the pungi.conf jinja2 template.",
        },
        "target_dir": {
            "type": str,
            "default": "/tmp",
            "desc": "Path to target dir to which store composes",
        },
        "target_dir_url": {
            "type": str,
            "default": "http://localhost/odcs",
            "desc": "Public facing URL to target_dir.",
        },
        "extra_target_dirs": {
            "type": dict,
            "default": {},
            "desc": "Extra target_dirs to optionally store the compose on "
            "instead of the conf.target_dir. Key is the name "
            "of volume, value if path to target_dir.",
        },
        "seconds_to_live": {
            "type": int,
            "default": 24 * 60 * 60,
            "desc": "Default number of seconds for which the compose is available.",
        },
        "max_seconds_to_live": {
            "type": int,
            "default": 72 * 60 * 60,
            "desc": "Max number of seconds for which the compose is available.",
        },
        "mbs_url": {
            "type": str,
            "default": "http://localhost/module-build-service",
            "desc": "URL to MSB API.",
        },
        "num_concurrent_pungi": {
            "type": int,
            "default": 2,
            "desc": "Number of concurrent Pungi processes.",
        },
        "allowed_source_types": {
            "type": list,
            "default": ["tag", "module", "build"],
            "desc": "Allowed source types.",
        },
        "allowed_flags": {
            "type": list,
            "default": COMPOSE_FLAGS.keys(),
            "desc": "Allowed compose flags.",
        },
        "allowed_arches": {
            "type": list,
            "default": ["x86_64"],
            "desc": "Allowed compose arches.",
        },
        "allowed_results": {
            "type": list,
            "default": COMPOSE_RESULTS,
            "desc": "Allowed compose results.",
        },
        "allowed_sources": {"type": list, "default": [""], "desc": "Allowed sources."},
        "allowed_compose_types": {
            "type": list,
            "default": [""],
            "desc": "Allowed compose types.",
        },
        "auth_ldap_server": {
            "type": str,
            "default": "",
            "desc": "Server URL to query user's groups.",
        },
        "auth_ldap_group_base": {
            "type": str,
            "default": "",
            "desc": "Group base to query user's groups from LDAP server.",
        },
        "allowed_clients": {
            "type": dict,
            "default": {"groups": {}, "users": {}},
            "desc": "Groups and users that are allowed to generate composes.",
        },
        "admins": {
            "type": dict,
            "default": {"groups": [], "users": []},
            "desc": "Admin groups and users.",
        },
        "auth_backend": {
            "type": str,
            "default": "",
            "desc": "Select which authentication backend is enabled and work "
            "with frond-end authentication together.",
        },
        "auth_openidc_userinfo_uri": {
            "type": str,
            "default": "",
            "desc": "UserInfo endpoint to get user information from FAS.",
        },
        "auth_openidc_required_scopes": {
            "type": list,
            "default": [],
            "desc": "Required scopes for submitting request to run new compose.",
        },
        "base_module_names": {
            "type": set,
            "default": set(["platform", "bootstrap"]),
            "desc": (
                "Set of module names which defines the product version "
                "(by their stream) of modules depending on them."
            ),
        },
        "messaging_backend": {
            "type": str,
            "default": "",
            "desc": "Messaging backend, rhmsg or fedora-messaging.",
        },
        "messaging_broker_urls": {
            "type": list,
            "default": [],
            "desc": "List of messaging broker URLs.",
        },
        "messaging_cert_file": {
            "type": str,
            "default": "",
            "desc": "Path to certificate file used to authenticate ODCS by broker.",
        },
        "messaging_key_file": {
            "type": str,
            "default": "",
            "desc": "Path to private key file used to authenticate ODCS by broker.",
        },
        "messaging_ca_cert": {
            "type": str,
            "default": "",
            "desc": "Path to trusted CA certificate bundle.",
        },
        "messaging_topic_prefix": {
            "type": str,
            "default": "",
            "desc": "Prefix for MESSAGING_TOPIC and INTERNAL_MESSAGING_TOPIC.",
        },
        "messaging_topic": {
            "type": str,
            "default": "",
            "desc": "Messaging topic to which messages are sent.",
        },
        "internal_messaging_topic": {
            "type": str,
            "default": "",
            "desc": "Messaging topic to which internal ODCS-only messages are sent.",
        },
        "oidc_base_namespace": {
            "type": str,
            "default": "https://pagure.io/odcs/",
            "desc": "Base namespace of OIDC scopes.",
        },
        "sigkeys": {
            "type": list,
            "default": [],
            "desc": "Default list of sigkeys. Any package in a compose must "
            "be signed by one of those keys. Can be overriden in a "
            "compose request.",
        },
        "pulp_server_url": {"type": str, "default": "", "desc": "Server URL of Pulp."},
        "pulp_username": {
            "type": str,
            "default": "",
            "desc": "Username to login Pulp.",
        },
        "pulp_password": {
            "type": str,
            "default": "",
            "desc": "Password to login Pulp.",
        },
        "koji_config": {"type": str, "default": None, "desc": "Koji config file."},
        "koji_profile": {"type": str, "default": None, "desc": "Koji config profile."},
        "koji_krb_ccache": {
            "type": str,
            "default": None,
            "desc": "Kerberos ccache file to use for Koji auth.",
        },
        "koji_krb_keytab": {
            "type": str,
            "default": None,
            "desc": "Kerberos keytab to use for Koji auth.",
        },
        "koji_krb_principal": {
            "type": str,
            "default": None,
            "desc": "Kerberos principal to use for Koji auth.",
        },
        "koji_tag_cache_cleanup_timeout": {
            "type": int,
            "default": 30,
            "desc": "Number of days after which the cached Koji tag data "
            'stored in the "koji_tag_cache" directory will be '
            "removed.",
        },
        "raw_config_urls": {
            "type": dict,
            "default": {},
            "desc": 'URLs to get the raw Pungi config from for "raw_config" '
            'source_type. Key is the name of the raw_config "source", '
            "value is the URL in which the %s string is replaced with, "
            '"commit_hash" value.',
        },
        "raw_config_wrapper_conf_path": {
            "type": str,
            "default": "/etc/odcs/raw_config_wrapper.conf",
            "desc": "Full path to the raw_config_wrapper.conf configuration "
            "file. This file holds Pungi configuration which should "
            "import real pungi configuration from raw_config.conf "
            "in order to override some variables.",
        },
        "raw_config_schema_override": {
            "type": str,
            "default": "",
            "desc": "Full path to the JSON file defining Pungi config schema "
            'override. This file is passed to "pungi-config-validate" '
            "using the --schema-override option and can influence "
            "the raw_config validation.",
        },
        "pungi_koji_args": {
            "type": list,
            "default": [],
            "desc": "Command line arguments used to construct pungi-koji " "command.",
        },
        "raw_config_pungi_koji_args": {
            "type": dict,
            "default": {},
            "desc": "Command line argument for raw_config source type, which "
            "overwrite arguments listed PUNGI_KOJI_ARGS.",
        },
        "celery_config": {
            "type": dict,
            "default": {},
            "desc": "Configuration dict to pass to Celery.",
        },
        "celery_broker_url": {"type": str, "default": "", "desc": "Celery broker URL"},
        "celery_pulp_composes_queue": {
            "type": str,
            "default": "pulp_composes",
            "desc": "Name of the Celery queue for Pulp composes.",
        },
        "celery_pungi_composes_queue": {
            "type": str,
            "default": "pungi_composes",
            "desc": "Name of the Celery queue for Pungi composes.",
        },
        "celery_cleanup_queue": {
            "type": str,
            "default": "cleanup",
            "desc": "Name of the Celery queue for cleanup task.",
        },
        "celery_router_config": {
            "type": dict,
            "default": {
                "routing_rules": {
                    "odcs.server.celery_tasks.generate_pungi_compose": {
                        "pungi_composes": {"source_type": 3},
                    },
                    "odcs.server.celery_tasks.generate_pulp_compose": {
                        "pulp_composes": {"source_type": 4},
                    },
                },
                "cleanup_task": "odcs.server.celery_tasks.run_cleanup",
                "default_queue": "pungi_composes",
            },
            "desc": "Configuration for custom celery router.",
        },
        "runroot_extra_mounts": {
            "type": list,
            "default": [],
            "desc": "Extra mountpoint directories for odcs-mock-runroot.",
        },
    }

    def __init__(self, conf_section_obj):
        """
        Initialize the Config object with defaults and then override them
        with runtime values.
        """

        # read items from conf and set
        for key in dir(conf_section_obj):
            # skip keys starting with underscore
            if key.startswith("_"):
                continue
            # set item (lower key)
            self.set_item(key.lower(), getattr(conf_section_obj, key))

        # set item from defaults if the item is not set
        for name, values in self._defaults.items():
            if hasattr(self, name):
                continue
            self.set_item(name, values["default"])

        # Used by Flask-Login to disable the @login_required decorator
        self.login_disabled = self.auth_backend == "noauth"

    def set_item(self, key, value):
        """
        Set value for configuration item. Creates the self._key = value
        attribute and self.key property to set/get/del the attribute.
        """
        if key == "set_item" or key.startswith("_"):
            raise Exception("Configuration item's name is not allowed: %s" % key)

        # Create the empty self._key attribute, so we can assign to it.
        setattr(self, "_" + key, None)

        # Create self.key property to access the self._key attribute.
        # Use the setifok_func if available for the attribute.
        setifok_func = "_setifok_{}".format(key)
        if hasattr(self, setifok_func):
            setx = lambda self, val: getattr(self, setifok_func)(val)
        else:
            setx = lambda self, val: setattr(self, "_" + key, val)
        getx = lambda self: getattr(self, "_" + key)
        delx = lambda self: delattr(self, "_" + key)
        setattr(Config, key, property(getx, setx, delx))

        # managed/registered configuration items
        if key in self._defaults:
            # type conversion for configuration item
            convert = self._defaults[key]["type"]
            if convert in [bool, int, list, str, set, dict, float]:
                try:
                    # Do no try to convert None...
                    if value is not None:
                        value = convert(value)
                except Exception:
                    raise TypeError(
                        "Configuration value conversion failed for name: %s" % key
                    )
            # unknown type/unsupported conversion
            elif convert is not None:
                raise TypeError(
                    "Unsupported type %s for configuration item name: %s"
                    % (convert, key)
                )

        # Set the attribute to the correct value
        setattr(self, key, value)

    #
    # Register your _setifok_* handlers here
    #

    def _setifok_allowed_clients(self, clients):
        if type(clients) != dict:
            raise TypeError("allowed_clients must be a dict.")
        for role, role_dict in clients.items():
            if role not in ["users", "groups"]:
                raise ValueError("Unknown role %s in allowed_clients." % role)
            if type(role_dict) != dict:
                raise TypeError("allowed_clients['%s'] is not a dict" % role)
            for user, user_dict in role_dict.items():
                if type(user_dict) != dict:
                    raise TypeError(
                        "allowed_clients['%s']['%s'] is not a dict" % (role, user)
                    )
                for key, value in user_dict.items():
                    if type(value) not in [set, list]:
                        raise ValueError(
                            "allowed_clients['%s']['%s']['%s'] is not a "
                            "list" % (role, user, key)
                        )
        self._allowed_clients = clients

    def _setifok_raw_config_urls(self, raw_config_urls):
        if type(raw_config_urls) != dict:
            raise TypeError("raw_config_urls must be a dict.")
        for name, url_data in raw_config_urls.items():
            if type(url_data) != dict:
                raise TypeError("raw_config_urls['%s'] is not a dict." % name)
            for key in ["url", "config_filename"]:
                if key not in url_data:
                    raise ValueError(
                        "raw_config_urls['%s']['%s'] is not defined." % (name, key)
                    )
        self._raw_config_urls = raw_config_urls

    def _setifok_log_backend(self, s):
        if s is None:
            self._log_backend = "console"
        elif s not in logger.supported_log_backends():
            raise ValueError("Unsupported log backend")
        self._log_backend = str(s)

    def _setifok_log_file(self, s):
        if s is None:
            self._log_file = ""
        else:
            self._log_file = str(s)

    def _setifok_log_level(self, s):
        level = str(s).lower()
        self._log_level = logger.str_to_log_level(level)

    def _setifok_target_dir(self, s):
        if not os.path.isabs(s):
            raise ValueError("Compose target dir is not an absolute path: %s" % s)
        if not (os.path.exists(s) and os.path.isdir(s)):
            raise ValueError(
                "Compose target dir doesn't exist or not a directory: %s" % s
            )
        self._target_dir = s

    def _setifok_pungi_conf_path(self, s):
        if not os.path.isfile(s):
            raise ValueError("Pungi config template doesn't exist: %s" % s)
        self._pungi_conf_path = s

    def _setifok_celery_router_config(self, celery_router_config):
        if type(celery_router_config) != dict:
            raise TypeError("celery_router_config must be a dict.")

        required_config_keys = ["routing_rules", "cleanup_task", "default_queue"]

        for conf_key in required_config_keys:
            if not celery_router_config.get(conf_key):
                raise KeyError("celery_router_config is missing %s" % conf_key)

        if type(celery_router_config["routing_rules"]) != dict:
            raise TypeError("routing_rules must be a dict.")

        self._celery_router_config = celery_router_config



"""Add compose_type, label and pungi_compose_id columns.

Revision ID: cd0781bbdab1
Revises: de0a86d7de49
Create Date: 2020-01-09 11:14:30.378081

"""

# revision identifiers, used by Alembic.
revision = "cd0781bbdab1"
down_revision = "de0a86d7de49"

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column("composes", sa.Column("compose_type", sa.String(), nullable=True))
    op.add_column("composes", sa.Column("label", sa.String(), nullable=True))
    op.add_column("composes", sa.Column("pungi_compose_id", sa.String(), nullable=True))


def downgrade():
    op.drop_column("composes", "pungi_compose_id")
    op.drop_column("composes", "label")
    op.drop_column("composes", "compose_type")



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>


from functools import wraps
import requests
import ldap
import flask

from itertools import chain

from flask import g
from flask_login import login_required as _login_required

from werkzeug.exceptions import Unauthorized
from odcs.server import conf, log
from odcs.server.errors import Forbidden
from odcs.server.models import User
from odcs.server.models import commit_on_success


def _validate_kerberos_config():
    """
    Validates the kerberos configuration and raises ValueError in case of
    error.
    """
    errors = []
    if not conf.auth_ldap_server:
        errors.append(
            "kerberos authentication enabled with no LDAP server configured, "
            "check AUTH_LDAP_SERVER in your config."
        )

    if not conf.auth_ldap_group_base:
        errors.append(
            "kerberos authentication enabled with no LDAP group base configured, "
            "check AUTH_LDAP_GROUP_BASE in your config."
        )

    if errors:
        for error in errors:
            log.exception(error)
        raise ValueError("Invalid configuration for kerberos authentication.")


@commit_on_success
def load_krb_user_from_request(request):
    """Load Kerberos user from current request

    REMOTE_USER needs to be set in environment variable, that is set by
    frontend Apache authentication module.
    """
    remote_user = request.environ.get("REMOTE_USER")
    if not remote_user:
        raise Unauthorized("REMOTE_USER is not present in request.")

    username, realm = remote_user.split("@")

    user = User.find_user_by_name(username)
    if not user:
        user = User.create_user(username=username)

    try:
        groups = query_ldap_groups(username)
    except ldap.SERVER_DOWN as e:
        log.error(
            "Cannot query groups of %s from LDAP. Error: %s",
            username,
            e.args[0]["desc"],
        )
        groups = []

    g.groups = groups
    g.user = user
    return user


@commit_on_success
def load_ssl_user_from_request(request):
    """
    Loads SSL user from current request.

    SSL_CLIENT_VERIFY and SSL_CLIENT_S_DN needs to be set in
    request.environ. This is set by frontend httpd mod_ssl module.
    """
    ssl_client_verify = request.environ.get("SSL_CLIENT_VERIFY")
    if ssl_client_verify != "SUCCESS":
        raise Unauthorized("Cannot verify client: %s" % ssl_client_verify)

    username = request.environ.get("SSL_CLIENT_S_DN")
    if not username:
        raise Unauthorized(
            "Unable to get user information (DN) from client certificate"
        )

    user = User.find_user_by_name(username)
    if not user:
        user = User.create_user(username=username)

    g.groups = []
    g.user = user
    return user


def load_krb_or_ssl_user_from_request(request):
    """
    Loads User using Kerberos or SSL auth.
    """
    if request.environ.get("REMOTE_USER"):
        return load_krb_user_from_request(request)
    else:
        return load_ssl_user_from_request(request)


def query_ldap_groups(uid):
    client = ldap.initialize(conf.auth_ldap_server)
    groups = client.search_s(
        conf.auth_ldap_group_base,
        ldap.SCOPE_ONELEVEL,
        attrlist=["cn", "gidNumber"],
        filterstr="memberUid={0}".format(uid),
    )

    group_names = list(chain(*[info["cn"] for _, info in groups]))
    return group_names


@commit_on_success
def load_openidc_user(request):
    """Load FAS user from current request"""
    username = request.environ.get("REMOTE_USER")
    if not username:
        raise Unauthorized("REMOTE_USER is not present in request.")

    token = request.environ.get("OIDC_access_token")
    if not token:
        raise Unauthorized("Missing token passed to ODCS.")

    scope = request.environ.get("OIDC_CLAIM_scope")
    if not scope:
        raise Unauthorized("Missing OIDC_CLAIM_scope.")
    validate_scopes(scope)

    user_info = get_user_info(token)

    user = User.find_user_by_name(username)
    if not user:
        user = User.create_user(username=username)

    g.groups = user_info.get("groups", [])
    g.user = user
    g.oidc_scopes = scope.split(" ")
    return user


def validate_scopes(scope):
    """Validate if request scopes are all in required scope

    :param str scope: scope passed in from.
    :raises: Unauthorized if any of required scopes is not present.
    """
    scopes = scope.split(" ")
    required_scopes = conf.auth_openidc_required_scopes
    for scope in required_scopes:
        if scope not in scopes:
            raise Unauthorized("Required OIDC scope {0} not present.".format(scope))


def require_oidc_scope(scope):
    """Check if required scopes is in OIDC scopes within request"""
    full_scope = "{0}{1}".format(conf.oidc_base_namespace, scope)
    if conf.auth_backend == "openidc" and full_scope not in g.oidc_scopes:
        message = "Request does not have required scope %s" % scope
        log.error(message)
        raise Forbidden(message)


def require_scopes(*scopes):
    """Check if required scopes is in OIDC scopes within request"""

    def wrapper(f):
        @wraps(f)
        def decorator(*args, **kwargs):
            if conf.auth_backend != "noauth":
                for scope in scopes:
                    require_oidc_scope(scope)
            return f(*args, **kwargs)

        return decorator

    return wrapper


def get_user_info(token):
    """Query FAS groups from Fedora"""
    headers = {"authorization": "Bearer {0}".format(token)}
    r = requests.get(
        conf.auth_openidc_userinfo_uri, headers=headers, timeout=conf.net_timeout
    )
    if r.status_code != 200:
        # In Fedora, the manually created service tokens can't be used with the UserInfo
        # endpoint. We treat this as an empty response - and hence an empty group list. An empty
        # group list only makes our authorization checks more strict, so it should be safe
        # to proceed and check the user.
        log.warning(
            "Failed to query group information - UserInfo endpoint failed with status=%d",
            r.status_code,
        )
        return {}

    return r.json()


def init_auth(login_manager, backend):
    """Initialize authentication backend

    Enable and initialize authentication backend to work with frontend
    authentication module running in Apache.
    """
    if backend == "noauth":
        # Do not enable any authentication backend working with frontend
        # authentication module in Apache.
        log.warning("Authorization is disabled in ODCS configuration.")
        return
    if backend == "kerberos":
        _validate_kerberos_config()
        global load_krb_user_from_request
        load_krb_user_from_request = login_manager.request_loader(
            load_krb_user_from_request
        )
    elif backend == "openidc":
        global load_openidc_user
        load_openidc_user = login_manager.request_loader(load_openidc_user)
    elif backend == "kerberos_or_ssl":
        _validate_kerberos_config()
        global load_krb_or_ssl_user_from_request
        load_krb_or_ssl_user_from_request = login_manager.request_loader(
            load_krb_or_ssl_user_from_request
        )
    elif backend == "ssl":
        global load_ssl_user_from_request
        load_ssl_user_from_request = login_manager.request_loader(
            load_ssl_user_from_request
        )
    else:
        raise ValueError("Unknown backend name {0}.".format(backend))


def has_role(role):
    """Check if current user has given role. With noauth all users have all
    roles.

    :returns: bool
    """
    if conf.auth_backend == "noauth":
        return True

    groups = []
    for group in getattr(conf, role).get("groups", []):
        groups.append(group)

    users = []
    for user in getattr(conf, role).get("users", []):
        users.append(user)

    in_groups = bool(set(flask.g.groups) & set(groups))
    in_users = flask.g.user.username in users
    if in_groups or in_users:
        return True
    return False


def requires_role(role):
    """Check if user is in the configured role.

    :param str role: role name, supported roles: 'allowed_clients', 'admins'.
    """
    valid_roles = ["allowed_clients", "admins"]
    if role not in valid_roles:
        raise ValueError(
            "Unknown role <%s> specified, supported roles: %s."
            % (role, str(valid_roles))
        )

    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            if has_role(role):
                return f(*args, **kwargs)

            msg = "User %s is not in role %s." % (flask.g.user.username, role)
            log.error(msg)
            raise Forbidden(msg)

        return wrapped

    return wrapper


def login_required(f):
    """Wrapper of flask_login's login_required to ingore auth check when auth backend is 'noauth'."""

    @wraps(f)
    def wrapped(*args, **kwargs):
        if conf.auth_backend == "noauth":
            return f(*args, **kwargs)
        return _login_required(f)(*args, **kwargs)

    return wrapped



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import os
import json
import requests
import time

from six.moves import urllib_parse
from requests_kerberos import HTTPKerberosAuth


class AuthMech(object):
    OpenIDC = 1
    Kerberos = 2
    Anonymous = 3
    SSL = 4

    @classmethod
    def has(cls, mech):
        return mech in (cls.OpenIDC, cls.Kerberos, cls.Anonymous, cls.SSL)


class ComposeLog(object):
    def __init__(self, compose):
        """
        Creates new ComposeLog instance.
        """
        self.url = os.path.join(compose["toplevel_url"], "pungi-stderr.log")
        self.offset = 0

    def read(self):
        """
        Reads the Compose log from the ODCS server and returns its content.

        This method can be called repeatedly to get the latest content in
        the log. Similar to "tail -f log".

        :return str: New log lines or None if log does not exist (yet) on the
            ODCS server.
        """
        headers = {"Range": "bytes=%d-" % self.offset}
        r = requests.get(self.url, headers=headers)

        # Log does not exists yet on the ODCS server.
        if r.status_code == 404:
            return None

        # 416 Range Not Satisfiable - nothing new in log.
        if r.status_code == 416:
            return ""

        r.raise_for_status()

        content = r.text
        self.offset += len(content)
        return content


def validate_int(value, min=1, type_error=None, value_error=None):
    if not isinstance(value, int):
        if type_error:
            raise TypeError(type_error)
        else:
            raise TypeError("Value {0} is not an integer.".format(value))
    if value < min:
        if value_error:
            raise ValueError(value_error)
        else:
            raise ValueError(
                "Value {0} is less than minimal value {1}.".format(value, min)
            )


def validate_page(value):
    validate_int(
        value,
        type_error="page number must be an integer.",
        value_error="page number must start from 1.",
    )


def validate_per_page(value):
    validate_int(
        value,
        type_error="per_page must be an integer.",
        value_error="per_page must be greater than or equal to 1.",
    )


class ComposeSourceGeneric(object):
    def __init__(
        self,
        source,
        source_type,
        packages=None,
        builds=None,
        sigkeys=None,
        koji_event=None,
        modular_koji_tags=None,
        module_defaults_url=None,
        module_defaults_commit=None,
        scratch_modules=None,
        **kwargs
    ):
        self.source = {"source": source, "type": source_type}
        if packages:
            self.source["packages"] = packages
        if builds:
            self.source["builds"] = builds
        if sigkeys:
            self.source["sigkeys"] = sigkeys
        if koji_event:
            self.source["koji_event"] = koji_event
        if modular_koji_tags:
            self.source["modular_koji_tags"] = modular_koji_tags
        if module_defaults_url:
            self.source["module_defaults_url"] = module_defaults_url
        if module_defaults_commit:
            self.source["module_defaults_commit"] = module_defaults_commit
        if scratch_modules:
            self.source["scratch_modules"] = scratch_modules
        self.source.update(kwargs)


class ComposeSourceTag(ComposeSourceGeneric):
    """
    Compose source taking Koji tag as input.
    """

    def __init__(
        self,
        tag,
        packages=None,
        builds=None,
        sigkeys=None,
        koji_event=None,
        modular_koji_tags=None,
        module_defaults_url=None,
        module_defaults_commit=None,
        scratch_modules=None,
        **kwargs
    ):
        """
        Creates new ComposeSourceTag instance.

        :param str tag: Koji tag to use as a source for compose.
        :param list packages: List of Koji packages to include in the compose. Note
            that this is **not** a list of RPM names. If unset, all packages tagged in
            the Koji ``tag`` will be included in a compose.
        :param list builds: List of NVRs of Koji builds to be included in the compose
            on top of the Koji builds directly tagged in the Koji ``tag``. If unset,
            only Koji builds tagged in the Koji ``tag`` will be used.
        :param list sigkeys: List of signature keys by which the RPMs in the compose must be
            signed. Empty string in the list allows unsigned packages.
        :param int koji_event: Particular Koji event id to generate compose from. If unset,
            latest Koji event will be used.
        :param list modular_koji_tags: List of Koji tags containing modules which should also
            be included in the resulting compose on top of Koji builds tagged in the Koji ``tag``.
        :param str module_defaults_url: URL of module defaults repository.
        :param str module_defaults_commit: Commit or branch name defining particular point
            module defaults repository.
        :param list scratch_modules: List of N:S:V:C of scratch modules to be included in
            the compose.
        """
        super(ComposeSourceTag, self).__init__(
            tag,
            "tag",
            packages,
            builds,
            sigkeys,
            koji_event,
            modular_koji_tags,
            module_defaults_url,
            module_defaults_commit,
            scratch_modules,
            **kwargs
        )


class ComposeSourceModule(ComposeSourceGeneric):
    """
    Compose source taking list of modules as input.
    """

    def __init__(
        self,
        modules,
        sigkeys=None,
        module_defaults_url=None,
        module_defaults_commit=None,
        scratch_modules=None,
        **kwargs
    ):
        """
        Creates new ComposeSourceModule instance.

        :param list modules: List of modules in N:S, N:S:V or N:S:V:C format.
        :param list sigkeys: List of signature keys by which the RPMs in the compose must be
            signed. Empty string in the list allows unsigned packages.
        :param str module_defaults_url: URL of module defaults repository.
        :param str module_defaults_commit: Commit or branch name defining particular point
            module defaults repository.
        :param list scratch_modules: List of N:S:V:C of scratch modules to be included in
            the compose.
        """
        super(ComposeSourceModule, self).__init__(
            " ".join(modules),
            "module",
            sigkeys=sigkeys,
            module_defaults_url=module_defaults_url,
            module_defaults_commit=module_defaults_commit,
            scratch_modules=scratch_modules,
            **kwargs
        )


class ComposeSourcePulp(ComposeSourceGeneric):
    """
    Compose source taking list of Pulp content_sets as input.
    """

    def __init__(self, content_sets, **kwargs):
        """
        Creates new ComposeSourcePulp instance.

        :param list content_sets: List of Pulp content-sets. Repositories defined by these
            content-sets will be included in the compose.
        """
        super(ComposeSourcePulp, self).__init__(
            " ".join(content_sets), "pulp", **kwargs
        )


class ComposeSourceRawConfig(ComposeSourceGeneric):
    """
    Compose source taking raw Pungi configuration file as input.
    """

    def __init__(self, config_name, commit, koji_event=None, **kwargs):
        """
        Creates new ComposeSourceRawConfig instance.

        :param str config_name: The name of Raw Pungi configuration as configured in ODCS server.
        :param str commit: The commit hash or branch to get the Pungi configuration from.
        :param int koji_event: Particular Koji event id to generate compose from. If unset,
            latest Koji event will be used.
        """
        super(ComposeSourceRawConfig, self).__init__(
            "%s#%s" % (config_name, commit),
            "raw_config",
            koji_event=koji_event,
            **kwargs
        )


class ComposeSourceBuild(ComposeSourceGeneric):
    """
    Compose source taking list of Koji builds as input.
    """

    def __init__(self, builds, sigkeys=None, **kwargs):
        """
        Creates new ComposeSourceModule instance.

        :param list builds: List of NVRs of Koji builds to be included in the compose.
        :param list sigkeys: List of signature keys by which the RPMs in the compose must be
            signed. Empty string in the list allows unsigned packages.
        """
        super(ComposeSourceBuild, self).__init__(
            "", "build", builds=builds, sigkeys=sigkeys, **kwargs
        )


class ODCS(object):
    """Client API to interact with ODCS APIs"""

    def __init__(
        self,
        server_url,
        api_version="1",
        verify_ssl=True,
        auth_mech=None,
        openidc_token=None,
        ssl_cert=None,
        ssl_key=None,
    ):
        """Initialize ODCS client

        :param str server_url: server URL of ODCS.
        :param str api_version: API version client will call. Version 1 is the default.
        :param bool verify_ssl: whether to verify SSL certificate over HTTP. By
            default, always verify, but you are also always able to disable it
            by passing False.
        :param AuthMech auth_mech: specify what authentication mechanism is
            used to interact with ODCS server. Choose one mechanism from
            AuthMech. Anonymous can be passed to force client not send
            any authentication information. If this parameter is omitted,
            same as Anonymous.
        :param str openidc_token: token got from OpenIDC so that client can be
            authenticated by ODCS server. This is only required if
            ``AuthMech.OpenIDC`` is passed to parameter ``auth_mech``.
        :param str ssl_cert: Path to SSL client certificate to use. This is
            only required if ``AuthMech.SSL`` is passed to parameter
            ``auth_mech``.
        :param str ssl_key: Path to SSL client key to use. This is
            only required if ``AuthMech.SSL`` is passed to parameter
            ``auth_mech``.
        """
        self._server_url = server_url
        self._api_version = api_version
        self._verify_ssl = verify_ssl
        if auth_mech == AuthMech.OpenIDC and not openidc_token:
            raise ValueError(
                "OpenIDC token must be specified when OpenIDC"
                " authentication is enabled."
            )
        self._openidc_token = openidc_token

        if auth_mech == AuthMech.SSL and (not ssl_cert or not ssl_key):
            raise ValueError(
                "SSL cert and keymust be specified when SSL"
                " authentication is enabled."
            )
        self._ssl_cert = ssl_cert
        self._ssl_key = ssl_key

        if auth_mech is None:
            self._auth_mech = AuthMech.Anonymous
        else:
            if not AuthMech.has(auth_mech):
                raise ValueError(
                    "Unknown authentication mechanism {0}".format(auth_mech)
                )
            self._auth_mech = auth_mech

    @property
    def server_url(self):
        return self._server_url

    @property
    def api_version(self):
        return self._api_version

    @property
    def auth_mech(self):
        return self._auth_mech

    def _make_endpoint(self, resource_path):
        """Helper method to construct URL to requested resource

        URL of requested resource consists of the server URL, API version and
        the resource path.

        :param str resource_path: the part after API version representing
            the concrete resource.
        :return: the whole complete URL of requested resource.
        :rtype: str
        """
        return urllib_parse.urljoin(
            self._server_url,
            "api/{0}/{1}".format(self.api_version, resource_path.lstrip("/")),
        )

    def _make_request(self, method, resource_path, data=None):
        """Make a HTTP request to server

        :param str method: HTTP request method to send, GET, POST and DELETE
            are supported.
        :param str resource_path: path of requested resource.
        :param dict data: corresponding data with specific request. It is
            optional. None is default that means no data is send along with
            request.
        :return: requests Response object.
        :rtype: requests.Response
        :raises: if ODCS does not response 200 or 202, exception will be raised
            by ``requests.Response.raise_for_status``.
        """
        request_data = {}
        headers = {}
        if data:
            if method in ("post", "patch"):
                request_data["data"] = json.dumps(data)
                headers["Content-Type"] = "application/json"
            if method == "get":
                request_data["params"] = data
        if not self._verify_ssl:
            request_data["verify"] = False
        if self.auth_mech == AuthMech.OpenIDC:
            headers["Authorization"] = "Bearer {0}".format(self._openidc_token)
        elif self.auth_mech == AuthMech.Kerberos:
            request_data["auth"] = HTTPKerberosAuth()
        elif self.auth_mech == AuthMech.SSL:
            request_data["cert"] = (self._ssl_cert, self._ssl_key)

        # Anonymous is the last possible value and no auth should be set

        if headers:
            request_data["headers"] = headers

        request_method = getattr(requests, method)
        resource_url = self._make_endpoint(resource_path)
        r = request_method(resource_url, **request_data)

        # Print error, for debugging
        if r.status_code not in (200, 202):
            print(r.text)

        r.raise_for_status()
        return r

    def _get(self, resource_path, data=None):
        """Make a GET HTTP request to server"""
        return self._make_request("get", resource_path, data)

    def _post(self, resource_path, data=None):
        """Make a POST HTTP request to server"""
        return self._make_request("post", resource_path, data)

    def _delete(self, resource_path, data=None):
        """Make a DELETE HTTP request to server"""
        return self._make_request("delete", resource_path, data)

    def _patch(self, resource_path, data=None):
        """Make a PATCH HTTP request to server"""
        return self._make_request("patch", resource_path, data)

    def new_compose(
        self,
        source,
        source_type,
        seconds_to_live=None,
        packages=[],
        flags=[],
        sigkeys=None,
        koji_event=None,
        results=None,
        arches=None,
        builds=None,
        modular_koji_tags=None,
        module_defaults_url=None,
        module_defaults_commit=None,
        scratch_modules=None,
        lookaside_repos=None,
        label=None,
        compose_type=None,
    ):
        """Request a new compose

        .. warning::
            This method is deprecated. Please use ``request_compose`` instead.

        :param str source: from where to grab and make new compose, different
            value for different ``source_type``. For ``tag`` source type, name
            of the tag. For ``module`` source type, white-space separated list
            of module name-stream or name-stream-version. For ``repo`` source
            type, full path to repository.
        :param str source_type: source type. ``tag`` for compose from Koji tag,
            ``module`` for compose from the Fedora module, ``repo`` for compose
            from local RPM repository.
        :param int seconds_to_live: Number of seconds for which the compose
            should become available.
        :param list packages: List of packages to include in a compose. Must not
            be set for "module" source_type.
        :param list flags: List of flags influencing the resulting compose.
            Valid flags are 1) ``no_deps``, the resulting compose will contain
            only packages defined in the "packages" list without their
            dependencies, or for ``source_type`` of "module", only the modules
            listed in ``source`` without their dependencies.
        :param list sigkeys: List of signature keys by which the packages
            in compose must be signed.
        :param int koji_event: Koji event for populating package set.
        :param list results: List of additional results which will be generated
            by ODCS as part of this compose. Can be "iso" for iso files with
            packages or "boot.iso" for images/boot.iso needed to generate
            container base images or installable DVDs.
        :return: the newly created Compose
        :rtype: dict
        """
        request_data = {"source": {"source": source, "type": source_type}}
        if packages:
            request_data["source"]["packages"] = packages
        if builds:
            request_data["source"]["builds"] = builds
        if sigkeys:
            request_data["source"]["sigkeys"] = sigkeys
        if koji_event:
            request_data["source"]["koji_event"] = koji_event
        if modular_koji_tags:
            request_data["source"]["modular_koji_tags"] = modular_koji_tags
        if module_defaults_url:
            request_data["source"]["module_defaults_url"] = module_defaults_url
        if module_defaults_commit:
            request_data["source"]["module_defaults_commit"] = module_defaults_commit
        if scratch_modules:
            request_data["source"]["scratch_modules"] = scratch_modules
        if lookaside_repos:
            request_data["lookaside_repos"] = lookaside_repos
        if label:
            request_data["label"] = label
        if compose_type:
            request_data["compose_type"] = compose_type
        if seconds_to_live is not None:
            request_data["seconds-to-live"] = seconds_to_live
        if flags:
            request_data["flags"] = flags
        if results:
            request_data["results"] = results
        if arches:
            request_data["arches"] = arches

        r = self._post("composes/", request_data)
        return r.json()

    def request_compose(self, source, **kwargs):
        """
        Request new compose.

        **Example**:

        .. sourcecode:: python

            source = ComposeSourceTag("f32-updates", packages="httpd")
            odcs = ODCS(...)
            compose = odcs.request_compose(source, arches="x86_64", flags=["nodeps"])
            odcs.wait_for_compose(compose["id"])

        :param ComposeSourceGeneric source: The source of compose.
        :param kwargs: Extra arguments passed to "/api/1/composes" POST JSON data.
        :return: the requested Compose object.
        :rtype: dict
        """
        request_data = {"source": source.source}
        request_data.update(kwargs)
        r = self._post("composes/", request_data)
        return r.json()

    def renew_compose(self, compose_id, seconds_to_live=None):
        """To regenerate an expired compose

        :param int compose_id: Compose ID to renew.
        :param int seconds_to_live: Number of seconds for which the compose
            should become available.
        :return: the new regenerated Compose
        :rtype: dict
        """
        if seconds_to_live is not None:
            request_data = {"seconds-to-live": seconds_to_live}
        else:
            request_data = None

        r = self._patch("composes/{0}".format(compose_id), request_data)
        return r.json()

    def find_composes(self, **search_criteria):
        """Find composes

        :param dict search_criteria: a mapping containing compose search
            criteria and pagination arguments. Composes can be searched
            by ``owner``, ``source_type``, ``source`` and ``state``.
        :return: list of found composes, each of them is a dict.
        :rtype: list
        """
        if "page" in search_criteria:
            validate_page(search_criteria["page"])
        if "per_page" in search_criteria:
            validate_per_page(search_criteria["per_page"])

        r = self._get("composes/", search_criteria)
        return r.json()

    def delete_compose(self, compose_id):
        """Delete a compose

        :param int compose_id: compose ID.
        :return: a mapping representing the acknowledge of a compose is delete.
        :rtype: dict
        """
        r = self._delete("composes/{0}".format(compose_id))
        return r.json()

    def get_compose(self, compose_id):
        """Get a compose

        :param int compose_id: compose ID.
        :return: a mapping representing a compose.
        :rtype: dict
        """
        r = self._get("composes/{0}".format(compose_id))
        return r.json()

    def wait_for_compose(self, compose_id, timeout=300, watch_logs=False):
        """
        Polls the ODCS server repeatedly to find out whether the compose
        moved from "wait" or "generating" state to some final state. Blocks
        for `timeout` seconds. If the compose does not move to the final state
        in given time limit, raises RuntimeError.

        NOTE: It is advised to use this method only in situations where the
        caller is not able to listen on Fedora messaging bus or UMB bus for the
        odcs.state.change messages which can inform the caller about the
        compose state change without the polling.

        :param int compose_id: compose ID.
        :param int timeout: Number of seconds to wait/block.
        :param bool watch_logs: If True, this method prints the compose log to
            stdout every 10 seconds while waiting for the compose to finish.
        :rtype: dict
        :return: a mapping representing a compose
        """
        elapsed = 0
        if watch_logs:
            sleep_time = 10
            compose = self.get_compose(compose_id)
            log = ComposeLog(compose)
        else:
            sleep_time = 1
            log = None
        start_time = time.time()
        while True:
            compose = self.get_compose(compose_id)
            if log:
                data = log.read()
                if data:
                    print(data)
            if compose["state_name"] not in ["wait", "generating"]:
                return compose

            elapsed = time.time() - start_time
            if elapsed >= timeout:
                raise RuntimeError(
                    "Retrieving ODCS compose %s timed out after %s seconds"
                    % (compose_id, timeout)
                )

            time.sleep(sleep_time)

            # Increase the sleep time for next try. But do not try sleeping
            # longer than the `timeout`.
            elapsed = time.time() - start_time

            # Do not increase sleep time in case we are watching logs.
            if not watch_logs:
                sleep_time = round(sleep_time * 1.5)

            if elapsed + sleep_time > timeout:
                sleep_time = timeout - elapsed



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import copy
import os
import shutil
import tempfile
import jinja2
import time
import threading
import json
from productmd.composeinfo import ComposeInfo
from kobo.conf import PyConfigParser

import odcs.server.utils
from odcs.server import conf, log, db
from odcs.server import comps
from odcs.server.models import Compose
from odcs.common.types import (
    PungiSourceType,
    COMPOSE_RESULTS,
    MULTILIB_METHODS,
    INVERSE_PUNGI_SOURCE_TYPE_NAMES,
    COMPOSE_FLAGS,
)
from odcs.server.utils import makedirs, clone_repo, copytree


class BasePungiConfig(object):
    def __init__(self):
        self.pungi_timeout = conf.pungi_timeout

    def _write(self, path, cfg):
        """
        Writes configuration string `cfg` to file defined by `path`.

        :param str path: Full path to file to write to.
        :param str cfg: Configuration to write.
        """
        with open(path, "w") as f:
            log.info("Writing %s configuration to %s.", os.path.basename(path), path)
            f.write(cfg)

    def write_config_files(self, topdir):
        """Write configuration into files"""
        raise NotImplementedError("Concrete config object must implement.")

    def validate(self, topdir, compose_dir):
        """Validate configuration. Raises an exception of error found."""
        pass


class RawPungiConfig(BasePungiConfig):
    def __init__(self, compose):
        super(RawPungiConfig, self).__init__()
        self.compose = compose
        source_name, source_hash = compose.source.split("#")

        url_data = copy.deepcopy(conf.raw_config_urls[source_name])
        # Do not override commit hash by hash from ODCS client if it is
        # hardcoded in the config file.
        if "commit" not in url_data:
            url_data["commit"] = source_hash

        self.pungi_timeout = url_data.get("pungi_timeout", conf.pungi_timeout)
        self.raw_config_wrapper_conf_path = url_data.get(
            "raw_config_wrapper", conf.raw_config_wrapper_conf_path
        )
        self.pungi_cfg = url_data
        self.pungi_koji_args = conf.raw_config_pungi_koji_args.get(
            source_name, conf.pungi_koji_args
        )

    def apply_raw_config_wrapper_overrides(self, cfg_path):
        try:
            with open(cfg_path) as fd:
                raw_config_wrapper = jinja2.Template(fd.read())
            with open(cfg_path, "w") as fd:
                fd.write(raw_config_wrapper.render(compose=self.compose.json()))
        except Exception as e:
            log.exception(
                "Failed to render raw_config_wrapper template {!r}: {}".format(
                    cfg_path, str(e)
                )
            )
            raise

    def write_config_files(self, topdir):
        """Write raw config files

        :param str topdir: Directory to write the files to.
        """
        # In case the raw_config wrapper config is set, download the
        # original pungi.conf as "raw_config.conf" and use
        # the raw_config wrapper as real "pungi.conf".
        # The reason is that wrapper config can import raw_config
        # and override some variables.
        if self.raw_config_wrapper_conf_path:
            main_cfg_path = os.path.join(topdir, "raw_config.conf")
            shutil.copyfile(
                self.raw_config_wrapper_conf_path, os.path.join(topdir, "pungi.conf")
            )
            self.apply_raw_config_wrapper_overrides(os.path.join(topdir, "pungi.conf"))
        else:
            main_cfg_path = os.path.join(topdir, "pungi.conf")

        # Clone the git repo with raw_config pungi config files.
        repo_dir = os.path.join(topdir, "raw_config_repo")
        clone_repo(self.pungi_cfg["url"], repo_dir, commit=self.pungi_cfg["commit"])

        # If the 'path' is defined, copy only the files form the 'path'
        # to topdir.
        if "path" in self.pungi_cfg:
            repo_dir = os.path.join(repo_dir, self.pungi_cfg["path"])

        copytree(repo_dir, topdir)

        # If pungi_config_dump is defined, it means that we are regenerating
        # already built production compose and this should use the very same
        # Pungi config file as the original one.
        # We still want to use the latest raw_config_wrapper, because this
        # is used to configure some important variables which can change
        # in the future ODCS deployment - like "koji_profile" or "runroot_method".
        # We also want to be able to respin the compose with the very same
        # configuration but include one new package build to fix last-minute
        # issues with the compose - this new build will also be set using the
        # raw_config_wrapper.
        if self.compose.pungi_config_dump:
            # Load the config dump.
            pungi_config_dump = json.loads(self.compose.pungi_config_dump)

            # If raw_config_wrapper is used, update the pungi_config_dump according
            # to it.
            if self.raw_config_wrapper_conf_path:
                # At first create empty "raw_config.conf" file so "raw_config_wrapper.conf"
                # can import it and can be loaded by PyConfigParser.
                with open(main_cfg_path, "w") as fd:
                    fd.write("# empty config file")

                # Load the raw_config_wrapper and update pungi_config_dump.
                raw_config_wrapper = PyConfigParser()
                raw_config_wrapper.load_from_file(main_cfg_path)
                pungi_config_dump.update(raw_config_wrapper)

            # Create the pungi.json
            with open(os.path.join(topdir, "pungi.json"), "w") as fd:
                fd.write(json.dumps(pungi_config_dump))
        else:
            # Create the "pungi.conf" from config_filename.
            config_path = os.path.join(topdir, self.pungi_cfg["config_filename"])
            if config_path != main_cfg_path:
                shutil.copy2(config_path, main_cfg_path)
            c = PyConfigParser()

            # Create pungi.json.
            c.load_from_file(os.path.join(topdir, "pungi.conf"))
            with open(os.path.join(topdir, "pungi.json"), "w") as fd:
                fd.write(json.dumps(c))

    def validate(self, topdir, compose_dir):
        if not conf.pungi_config_validate:
            return
        pungi_config_validate_cmd = [conf.pungi_config_validate, "--old-composes"]

        # Apply global schema override.
        if conf.raw_config_schema_override:
            pungi_config_validate_cmd += [
                "--schema-override",
                conf.raw_config_schema_override,
            ]

        # Apply raw_config specific schema override.
        if "schema_override" in self.pungi_cfg:
            pungi_config_validate_cmd += [
                "--schema-override",
                self.pungi_cfg["schema_override"],
            ]

        # Add raw_config configuration file to validate.
        pungi_config_validate_cmd.append(os.path.join(topdir, "pungi.json"))

        # Run the pungi-config-validate. The execute_cmd raises an exception
        # if config is invalid.
        log_out_path = os.path.join(compose_dir, "pungi-config-validate-stdout.log")
        log_err_path = os.path.join(compose_dir, "pungi-config-validate-stderr.log")
        with open(log_out_path, "w") as log_out:
            with open(log_err_path, "w") as log_err:
                odcs.server.utils.execute_cmd(
                    pungi_config_validate_cmd, stdout=log_out, stderr=log_err
                )


class PungiConfig(BasePungiConfig):
    def __init__(
        self,
        release_name,
        release_version,
        source_type,
        source,
        packages=None,
        arches=None,
        sigkeys=None,
        results=0,
        multilib_arches=None,
        multilib_method=0,
        builds=None,
        flags=0,
        lookaside_repos=None,
        modular_koji_tags=None,
        module_defaults_url=None,
        scratch_modules=None,
        scratch_build_tasks=None,
    ):
        super(PungiConfig, self).__init__()
        self.release_name = release_name
        self.release_version = release_version
        self.bootable = False
        self.sigkeys = sigkeys.split(" ") if sigkeys else []
        self.source_type = source_type
        self.source = source
        self.koji_profile = conf.koji_profile
        self.pkgset_koji_inherit = True
        self.lookaside_repos = lookaside_repos.split(" ") if lookaside_repos else []
        if arches:
            self.arches = arches
        else:
            self.arches = conf.arches
        self.packages = packages or []
        self.builds = builds or []
        self.scratch_modules = scratch_modules.split(" ") if scratch_modules else []
        self.scratch_build_tasks = (
            scratch_build_tasks.split(" ") if scratch_build_tasks else []
        )
        self.pkgset_allow_reuse = True

        # Store results as list of strings, so it can be used by jinja2
        # templates.
        self.results = []
        for k, v in COMPOSE_RESULTS.items():
            if results & v:
                self.results.append(k)

        self.multilib_arches = multilib_arches if multilib_arches else []
        self.multilib_method = []
        if multilib_method:
            for k, v in MULTILIB_METHODS.items():
                if multilib_method & v:
                    self.multilib_method.append(k)

        if "boot.iso" in self.results:
            self.bootable = True

        if source_type == PungiSourceType.KOJI_TAG:
            self.koji_module_tags = (
                modular_koji_tags.split(" ") if modular_koji_tags else []
            )
            self.module_defaults_url = (
                module_defaults_url.split(" ") if module_defaults_url else []
            )
            self.koji_tag = source
            self.gather_source = "comps"
            if self.koji_module_tags:
                self.gather_method = "hybrid"
            else:
                self.gather_method = "deps"
        elif source_type == PungiSourceType.MODULE:
            self.koji_tag = None
            self.gather_source = "module"
            self.gather_method = "nodeps"
            self.module_defaults_url = (
                module_defaults_url.split(" ") if module_defaults_url else []
            )

            if self.packages:
                raise ValueError(
                    "Exact packages cannot be set for MODULE " "source type."
                )
        elif source_type in [
            PungiSourceType.BUILD,
            PungiSourceType.PUNGI_COMPOSE,
            PungiSourceType.REPO,
        ]:
            self.gather_source = "comps"
            self.gather_method = "deps"
            self.koji_tag = None
        else:
            raise ValueError("Unknown source_type %r" % source_type)

        self.check_deps = bool(flags & COMPOSE_FLAGS["check_deps"])

    @property
    def source_type_str(self):
        return INVERSE_PUNGI_SOURCE_TYPE_NAMES[self.source_type]

    @property
    def release_short(self):
        return self.release_name[:16]

    @property
    def comps_file(self):
        if self.source_type == PungiSourceType.MODULE:
            return None
        else:
            return "comps.xml"

    @property
    def pkgset_source(self):
        if self.source_type == PungiSourceType.REPO:
            return "repos"
        return "koji"

    @property
    def mbs_api_url(self):
        return os.path.join(conf.mbs_url, "1")

    def get_comps_config(self):
        if self.source_type == PungiSourceType.MODULE:
            return ""
        odcs_comps = comps.Comps()
        odcs_group = comps.Group(
            "odcs-group", "odcs-group", "ODCS compose default group"
        )
        for package in self.packages:
            odcs_group.add_package(comps.Package(package))
        odcs_comps.add_group(odcs_group)

        template = jinja2.Template(comps.COMPS_TEMPLATE)
        return template.render(comps=odcs_comps)

    def get_variants_config(self):
        odcs_product = comps.Product()
        tmp_variant = comps.Variant(
            "Temporary", "Temporary", "variant", self.source_type
        )
        for arch in self.arches:
            tmp_variant.add_arch(comps.Arch(arch))
        if self.source_type == PungiSourceType.MODULE:
            for module in self.source.split(" "):
                if module:
                    tmp_variant.add_module(comps.Module(module))
        elif self.source_type == PungiSourceType.KOJI_TAG:
            if self.packages:
                tmp_variant.add_group(
                    comps.Group(
                        "odcs-group", "odcs-group", "ODCS compose default group"
                    )
                )
            if self.koji_module_tags:
                tmp_variant.add_module(comps.Module("*"))

        odcs_product.add_variant(tmp_variant)

        template = jinja2.Template(comps.VARIANTS_TEMPLATE)
        return template.render(product=odcs_product)

    def get_pungi_config(self):
        try:
            with open(conf.pungi_conf_path) as fd:
                template = jinja2.Template(fd.read())
            c_data = template.render(config=self)
            c = PyConfigParser()
            c.load_from_string(c_data)
            return c
        except Exception as e:
            log.exception(
                "Failed to render pungi conf template {!r}: {}".format(
                    conf.pungi_conf_path, str(e)
                )
            )

    def write_config_files(self, topdir):
        """
        Writes "pungi.conf", "variants.xml" and "comps.xml" defined in
        `self.pungi_cfg` to `topdir` directory.

        :param str topdir: Directory to write the files to.
        """
        main_cfg = self.get_pungi_config()
        variants_cfg = self.get_variants_config()
        comps_cfg = self.get_comps_config()
        log.debug("Main Pungi config:")
        log.debug("%s", main_cfg)
        log.debug("Variants.xml:")
        log.debug("%s", variants_cfg)
        log.debug("Comps.xml:")
        log.debug("%s", comps_cfg)

        self._write(os.path.join(topdir, "pungi.json"), json.dumps(main_cfg))
        self._write(os.path.join(topdir, "variants.xml"), variants_cfg)
        self._write(os.path.join(topdir, "comps.xml"), comps_cfg)


class ReadComposeIdThread(threading.Thread):
    def __init__(self, compose):
        threading.Thread.__init__(self)
        self._stop_event = threading.Event()
        self.compose = compose

    def stop(self):
        self._stop_event.set()

    def run(self):
        p = os.path.join(
            self.compose.toplevel_dir, "work", "global", "composeinfo-base.json"
        )
        while not self._stop_event.is_set():
            time.sleep(1)

            # File does not exist yet.
            if not os.path.exists(p):
                continue

            ci = ComposeInfo()
            try:
                ci.load(p)
            except Exception:
                # This should happen only if file exists, but the data is not
                # written yet.
                continue

            self.compose.pungi_compose_id = ci.compose.id
            db.session.commit()
            break


class Pungi(object):
    def __init__(self, compose_id, pungi_cfg, koji_event=None, old_compose=None):
        self.compose_id = compose_id
        self.pungi_cfg = pungi_cfg
        self.koji_event = koji_event
        self.old_compose = old_compose

    def _write_cfgs(self, topdir):
        """Wrtie pungi config

        :param str topdir: Directory to write the files to.
        """
        self.pungi_cfg.write_config_files(topdir)

    def get_pungi_cmd(self, conf_topdir, compose, compose_dir=None):
        """
        Returns list with pungi command line arguments needed to generate
        the compose.
        :param str conf_topdir: Directory in which to look for Pungi
            configuration files.
        :param str targetdir: Target directory in which the compose should be
            generated.
        :param str compose_dir: If defined, overrides the Pungi compose_dir.
        :rtype: list
        :return: List of pungi command line arguments.
        """
        pungi_cmd = [
            conf.pungi_koji,
            "--config=%s" % os.path.join(conf_topdir, "pungi.json"),
            "--no-latest-link",
        ]

        if compose_dir:
            pungi_cmd.append("--compose-dir=%s" % compose_dir)
        else:
            pungi_cmd.append("--target-dir=%s" % compose.target_dir)

        if isinstance(self.pungi_cfg, RawPungiConfig):
            pungi_cmd += self.pungi_cfg.pungi_koji_args
        elif isinstance(self.pungi_cfg, PungiConfig):
            pungi_cmd += conf.pungi_koji_args
        else:
            raise RuntimeError("Unknown pungi config type to handle.")

        compose_type_to_arg = {
            "test": "--test",
            "ci": "--ci",
            "nightly": "--nightly",
            "production": "--production",
            "development": "--development",
        }
        compose_type = compose.compose_type or "test"

        # Add compose_type arg to pungi_cmd only if it's not set already
        # directly in the configuration.
        if not set(pungi_cmd).intersection(set(compose_type_to_arg.values())):
            # For unknown compose_type, fallback to --test to be safe.
            pungi_cmd.append(compose_type_to_arg.get(compose_type, "--test"))

        if compose.label:
            pungi_cmd.append("--label=%s" % str(compose.label))
        if compose.parent_pungi_compose_ids:
            for compose_id in compose.parent_pungi_compose_ids.split(" "):
                pungi_cmd.append("--parent-compose-id=%s" % str(compose_id))

        if self.koji_event:
            pungi_cmd += ["--koji-event", str(self.koji_event)]
        if self.old_compose:
            pungi_cmd += ["--old-composes", self.old_compose]
        return pungi_cmd

    def _prepare_compose_dir(self, compose, conf):
        """
        Creates the compose directory and returns the full path to it.
        """
        compose_date = time.strftime("%Y%m%d", time.localtime())
        makedirs(compose.toplevel_dir)

        # If Compose Tracking Service is configured in the config file,
        # we skip the ComposeInfo creation completely and instead let
        # the Pungi to ask CTS to generate unique ComposeInfo.
        if conf.get("cts_url") and conf.get("cts_keytab"):
            return compose.toplevel_dir

        ci = ComposeInfo()
        ci.release.name = conf["release_name"]
        ci.release.short = conf["release_short"]
        ci.release.version = conf["release_version"]
        ci.release.is_layered = True if conf.get("base_product_name", "") else False
        ci.release.type = conf.get("release_type", "ga").lower()
        ci.release.internal = bool(conf.get("release_internal", False))
        if ci.release.is_layered:
            ci.base_product.name = conf["base_product_name"]
            ci.base_product.short = conf["base_product_short"]
            ci.base_product.version = conf["base_product_version"]
            ci.base_product.type = conf.get("base_product_type", "ga").lower()

        ci.compose.label = compose.label
        ci.compose.type = compose.compose_type or "test"
        ci.compose.date = compose_date
        ci.compose.respin = 0

        while True:
            ci.compose.id = ci.create_compose_id()
            existing_compose = Compose.query.filter(
                Compose.pungi_compose_id == ci.compose.id
            ).first()
            if not existing_compose:
                break
            ci.compose.respin += 1

        # Dump the compose info to work/global/composeinfo-base.json.
        work_dir = os.path.join(compose.toplevel_dir, "work", "global")
        makedirs(work_dir)
        ci.dump(os.path.join(work_dir, "composeinfo-base.json"))

        compose.pungi_compose_id = ci.compose.id

        return compose.toplevel_dir

    def run_locally(self, compose):
        """
        Runs local Pungi compose.
        """
        td = None
        compose_id_thread = None
        try:
            td = tempfile.mkdtemp()
            self._write_cfgs(td)

            # Load pungi configuration file.
            with open(os.path.join(td, "pungi.json")) as fd:
                pungi_conf = json.load(fd)

            compose_dir = self._prepare_compose_dir(compose, pungi_conf)
            self.pungi_cfg.validate(td, compose_dir)
            pungi_cmd = self.get_pungi_cmd(td, compose, compose_dir)

            # Commit the session to ensure that all the `compose` changes are
            # stored database before executing the compose and are not just
            # cached locally in the SQLAlchemy.
            db.session.commit()

            # If Compose Tracking Service is configured in the config file,
            # we need to get the Compose ID from Pungi in separate thread.
            if pungi_conf.get("cts_url") and pungi_conf.get("cts_keytab"):
                compose_id_thread = ReadComposeIdThread(compose)
                compose_id_thread.start()

            log_out_path = os.path.join(compose_dir, "pungi-stdout.log")
            log_err_path = os.path.join(compose_dir, "pungi-stderr.log")

            with open(log_out_path, "w") as log_out:
                with open(log_err_path, "w") as log_err:
                    odcs.server.utils.execute_cmd(
                        pungi_cmd,
                        cwd=td,
                        timeout=self.pungi_cfg.pungi_timeout,
                        stdout=log_out,
                        stderr=log_err,
                    )
        finally:
            if compose_id_thread:
                compose_id_thread.stop()
            try:
                if td is not None:
                    shutil.rmtree(td)
            except Exception as e:
                log.warning(
                    "Failed to remove temporary directory {!r}: {}".format(td, str(e))
                )

    def run(self, compose):
        """
        Runs the compose in Pungi. Blocks until the compose is done.
        Raises an exception if compose generation fails.

        :param models.Compose compose: Compose this Pungi process is running
            for.
        """
        self.run_locally(compose)


class PungiLogs(object):
    def __init__(self, compose):
        self.compose = compose

    @property
    def global_log_path(self):
        """
        Returns the path to pungi.global.log if it exists.
        """
        toplevel_dir = self.compose.toplevel_dir
        if not toplevel_dir:
            return None
        return os.path.join(toplevel_dir, "logs", "global", "pungi.global.log")

    @property
    def config_dump_path(self):
        """
        Returns path to Pungi config dump.
        """
        toplevel_dir = self.compose.toplevel_dir
        if not toplevel_dir:
            return None
        return os.path.join(toplevel_dir, "logs", "global", "config-dump.global.log")

    def _get_global_log_errors(self):
        """
        Helper method which opens the `self.global_log_path` and search for
        all errors in that log file.

        :rtype: list
        :return: List of error strings.
        """
        errors = []
        global_log_path = self.global_log_path
        if not global_log_path:
            return errors
        try:
            with open(global_log_path, "r") as global_log:
                error = ""
                for line in global_log:
                    idx = line.find("[ERROR   ]")
                    if idx == -1:
                        if error:
                            error += line
                            errors.append(error)
                            error = ""
                        continue
                    if error:
                        errors.append(error)
                    error = line[idx + len("[ERROR   ] ") :]
        except IOError:
            pass
        return errors

    def get_error_string(self):
        """
        Returns the string with errors parsed from Pungi logs.

        :rtype: str
        :return: String with errors parsed from Pungi logs.
        """
        errors = ""

        global_errors = self._get_global_log_errors()
        for error in global_errors:
            if error.startswith("Extended traceback in:"):
                continue
            errors += error
            if len(errors) > 2000:
                errors = errors[:2000]
                errors += " ...\n Too many errors, see Pungi log for more information."
                break

        if self.compose.on_default_target_dir:
            errors = errors.replace(conf.target_dir, conf.target_dir_url)
        return errors

    def get_config_dump(self):
        config_dump_path = self.config_dump_path
        if not config_dump_path:
            return None

        with open(config_dump_path, "r") as config_dump:
            return config_dump.read()



# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

import six
import unittest
from mock import patch

from odcs.server.pungi_compose import PungiCompose


RPMS_JSON = {
    "header": {"type": "productmd.rpms", "version": "1.2"},
    "payload": {
        "compose": {
            "date": "20181210",
            "id": "odcs-691-1-20181210.n.0",
            "respin": 0,
            "type": "nightly",
        },
        "rpms": {
            "Temporary": {
                "x86_64": {
                    "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.src": {
                        "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.src": {
                            "category": "source",
                            "path": "Temporary/source/tree/Packages/f/flatpak-rpm-macros-29-6.module+125+c4f5c7f2.src.rpm",
                            "sigkey": None,
                        },
                        "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.x86_64": {
                            "category": "binary",
                            "path": "Temporary/x86_64/os/Packages/f/flatpak-rpm-macros-29-6.module+125+c4f5c7f2.x86_64.rpm",
                            "sigkey": None,
                        },
                    },
                    "flatpak-runtime-config-0:29-4.module+125+c4f5c7f2.src": {
                        "flatpak-runtime-config-0:29-4.module+125+c4f5c7f2.src": {
                            "category": "source",
                            "path": "Temporary/source/tree/Packages/f/flatpak-runtime-config-29-4.module+125+c4f5c7f2.src.rpm",
                            "sigkey": "sigkey1",
                        },
                        "flatpak-runtime-config-0:29-4.module+125+c4f5c7f2.x86_64": {
                            "category": "binary",
                            "path": "Temporary/x86_64/os/Packages/f/flatpak-runtime-config-29-4.module+125+c4f5c7f2.x86_64.rpm",
                            "sigkey": "sigkey1",
                        },
                    },
                }
            }
        },
    },
}


@patch("odcs.server.pungi_compose.PungiCompose._fetch_json")
class TestPungiCompose(unittest.TestCase):
    def test_get_rpms_data(self, fetch_json):
        fetch_json.return_value = RPMS_JSON
        compose = PungiCompose("http://localhost/compose/Temporary")
        data = compose.get_rpms_data()

        expected = {
            "sigkeys": set(["sigkey1", None]),
            "arches": set(["x86_64"]),
            "builds": {
                "flatpak-rpm-macros-29-6.module+125+c4f5c7f2": set(
                    [
                        "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.src",
                        "flatpak-rpm-macros-0:29-6.module+125+c4f5c7f2.x86_64",
                    ]
                ),
                "flatpak-runtime-config-29-4.module+125+c4f5c7f2": set(
                    [
                        "flatpak-runtime-config-0:29-4.module+125+c4f5c7f2.src",
                        "flatpak-runtime-config-0:29-4.module+125+c4f5c7f2.x86_64",
                    ]
                ),
            },
        }

        self.assertEqual(data, expected)

    def test_get_rpms_data_unknown_variant(self, fetch_json):
        fetch_json.return_value = RPMS_JSON
        msg = (
            "The http://localhost/compose/metadata/rpms.json does not "
            "contain payload -> rpms -> Workstation section"
        )
        with six.assertRaisesRegex(self, ValueError, msg):
            compose = PungiCompose("http://localhost/compose/Workstation")
            compose.get_rpms_data()



# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

from odcs.server.backend import BackendThread
from .utils import ModelsBaseTest
from mock import patch


class TestBackendThread(ModelsBaseTest):
    maxDiff = None

    def setUp(self):
        super(TestBackendThread, self).setUp()

        self.patch_do_work = patch(
            "odcs.server.backend.BackendThread.do_work", autospec=True
        )
        self.do_work = self.patch_do_work.start()

        self.thread = BackendThread()

    def tearDown(self):
        super(TestBackendThread, self).tearDown()

        self.patch_do_work.stop()

    @patch("odcs.server.backend.db.session.rollback")
    def test_do_work_exception(self, rollback):
        def mocked_do_work(backend_thread):
            # Stop the backend thread at first, so we won't wait for timeout.
            backend_thread.stop()
            raise ValueError("expected exception")

        self.do_work.side_effect = mocked_do_work
        self.thread._run()
        rollback.assert_called_once()
